# -*- coding: UTF-8 -*-
from insights.core import XMLParser
from insights.tests import context_wrap


class FakeXmlParserClass(XMLParser):
    def parse_dom(self):
        # default namespace
        xmlns = 'http://people.example.com'
        keyword = '*{%s}%s' % (xmlns, "neighbor[2]")
        tmo_dict = {}
        for n in self.dom.findall(keyword):
            tmo_dict = {n.get("name"): n.get("direction")}
        return tmo_dict

    @property
    def get_neighbors(self):
        return self.get_elements("./country/neighbor")


testdata = """
<?xml version="1.0"?>
<data xmlns:fictional="http://characters.example.com"
       xmlns="http://people.example.com">
    <country name="Liechtenstein">
        <rank updated="yes">2</rank>
        <year>2008</year>
        <gdppc>141100</gdppc>
        <neighbor name="Austria" direction="E"/>
        <neighbor name="Switzerland" direction="W"/>
    </country>
    <country name="Singapore">
        <rank updated="yes">5</rank>
        <year>2011</year>
        <gdppc>59900</gdppc>
        <neighbor name="Malaysia" direction="N"/>
    </country>
    <country name="Panama">
        <rank>68</rank>
        <year>2011</year>
        <gdppc>13600</gdppc>
        <neighbor name="Costa Rica" direction="W"/>
    </country>
</data>
""".strip()


def test_parse():
    ctx = context_wrap(testdata)
    xml = FakeXmlParserClass(ctx)
    assert xml.raw == testdata
    assert len(xml.get_neighbors) == 4
    assert xml.xmlns == "http://people.example.com"
    assert "Costa Rica" not in xml
    assert "Switzerland" in xml
    assert xml.get("Switzerland", "") == "W"
    assert len(xml.get_elements(".//year/..[@name='Singapore']")) == 1
    assert xml.get_elements(".//*[@name='Singapore']/year")[-1].text == '2011'
    assert xml.get_elements(".//neighbor[2]", xmlns="http://people.example.com")[0].get('name') == 'Switzerland'



"""
Hosts - file ``/etc/hosts``
===========================

This parser parses the ``/etc/hosts`` file, strips the comments, ignores the
blank lines, and collects the host names by IP address.  IPv4 and IPv6
addresses are supported.

Sample hosts file::

    127.0.0.1 localhost localhost.localdomain localhost4 localhost4.localdomain4
    ::1 localhost localhost.localdomain localhost6 localhost6.localdomain6
    # The same IP address can appear more than once, with different names
    127.0.0.1 fte.example.com

    10.0.0.1 nonlocal.example.com nonlocal2.fte.example.com
    10.0.0.2 other.host.example.com # Comments at end of line are ignored

Examples:

    >>> hosts = shared[Hosts]
    >>> len(hosts.all_names)
    10
    >>> hosts.all_names[0] # uses a set, so may not be in same order
    'localhost6'
    >>> hosts.data['127.0.0.1']
    ['localhost', 'localhost.localdomain', 'localhost4', 'localhost4.localdomain4',
     'fte.example.com']
    >>> hosts.get_nonlocal()
    {'10.0.0.1': ['nonlocal.example.com', 'nonlocal2.fte.example.com'],
     '10.0.0.2': ['other.host.example.com']}

"""

from collections import defaultdict
from .. import Parser, parser
from insights.specs import Specs


@parser(Specs.hosts)
class Hosts(Parser):
    """
    Read the ``/etc/hosts`` file and parse it into a dictionary of host name
    lists, keyed on IP address.
    """

    def parse_content(self, content):
        host_data = defaultdict(list)
        for line in content:
            line = line.strip()
            if "#" in line:
                line = line.split("#")[0]
            words = line.split(None)
            if len(words) > 1:
                host_data[words[0]].extend(words[1:])
        self.data = dict(host_data)

    @property
    def all_names(self):
        """
        (set): The set of host names known, regardless of their IP address.
        """
        names = set()
        for host_set in self.data.values():
            names.update(host_set)
        return names

    def get_nonlocal(self):
        """
        A dictionary of host name lists, keyed on IP address, that are not
        the 'localhost' addresses '127.0.0.1' or '::1'.
        """
        return dict((ip, host_list)
                    for ip, host_list in self.data.items()
                    if ip not in ("127.0.0.1", "::1"))

    def ip_of(self, hostname):
        """
        Return the (first) IP address given for this host name.  None is
        returned if no IP address is found.
        """
        for ip, host_list in self.data.items():
            if hostname in host_list:
                return ip
        return None



from insights.parsers.dnsmasq_config import DnsmasqConf
from insights.tests import context_wrap

DNSMASQ_CONF_MAIN = """
# Listen on this specific port instead of the standard DNS port
# (53). Setting this to zero completely disables DNS function,
# leaving only DHCP and/or TFTP.
port=5353

no-resolv
domain-needed
no-negcache
max-cache-ttl=1
enable-dbus
dns-forward-max=5000
cache-size=5000
bind-dynamic
except-interface=lo
server=/in-addr.arpa/127.0.0.1
server=/cluster.local/127.0.0.1
# End of config
""".strip()

DNSMASQ_CONF_FILE_1 = """
server=/in-addr.arpa/127.0.0.1
server=/cluster.local/127.0.0.1
""".strip()


def test_dnsmasq_conf():
    result = DnsmasqConf(context_wrap(DNSMASQ_CONF_MAIN, path="/etc/dnsmasq.conf"))
    assert "no-resolv" in result
    assert result.find("port").value == 5353
    assert len(result.find_all("server")) == 2
    assert result.find_all("server")[0].value == '/in-addr.arpa/127.0.0.1'
    assert result.find("bind-dynamic").name == 'bind-dynamic'
    assert "# End of config" not in result


def test_dnsmasq_conf_file():
    result = DnsmasqConf(context_wrap(DNSMASQ_CONF_FILE_1, path="/etc/dnsmasq.d/dns-origin.conf"))
    assert len(result["server"]) == 2
    assert result["server"][-1].value == "/cluster.local/127.0.0.1"



"""
GrubConf - The valid GRUB configuration
=======================================
Combiner for Red Hat Grub v1 and Grub v2 information.

This combiner uses the :class:`insights.parsers.grub_conf.Grub1Config`,
:class:`insights.parsers.grub_conf.Grub1EFIConfig`,
:class:`insights.parsers.grub_conf.Grub2Config` and
:class:`insights.parsers.grub_conf.Grub2EFIConfig` parsers.
It determines which parser was used by checking the InstalledRpms or CmdLine and
LsSysFirmware parsers.

Examples:
    >>> type(grub_conf)
    <class 'insights.combiners.grub_conf.GrubConf'>
    >>> grub_conf.version
    2
    >>> grub_conf.is_efi
    False
    >>> grub_conf.kernel_initrds
    {'grub_initrds': ['/initramfs-3.10.0-327.36.3.el7.x86_64.img'],
     'grub_kernels': ['/vmlinuz-3.10.0-327.36.3.el7.x86_64']}
    >>> grub_conf.is_kdump_iommu_enabled
    False
    >>> grub_conf.get_grub_cmdlines('')
    []
"""

from .. import defaults
from insights.core.plugins import combiner
from insights.parsers.grub_conf import Grub1Config, Grub1EFIConfig, Grub2Config, Grub2EFIConfig
from insights.parsers.ls_sys_firmware import LsSysFirmware
from insights.parsers.installed_rpms import InstalledRpms
from insights.parsers.cmdline import CmdLine
from insights.parsers import ParseException


@combiner([Grub1Config, Grub2Config, Grub1EFIConfig, Grub2EFIConfig],
          optional=[InstalledRpms, CmdLine, LsSysFirmware])
class GrubConf(object):
    """
    Process Grub configuration v1 or v2 based on which type is passed in.

    Raises:
        ParseException: when cannot find any valid grub configuration for the
                        booted system.

    Attributes:
        version (int): returns 1 or 2, version of the GRUB configuration
        is_efi (bool): returns true if grub configuration for EFI-based system
        kernel_initrds (dict): returns a dict of the `kernel` and `initrd`
            files referenced in GRUB configuration files
        is_kdump_iommu_enabled (bool): returns true if any kernel entry
            contains "intel_iommu=on"
    """

    def __init__(self, grub1, grub2, grub1_efi, grub2_efi,
                 rpms, cmdline, sys_firmware):

        self.version = self.is_kdump_iommu_enabled = None
        self.grub = self.kernel_initrds = None
        _grubs = list(filter(None, [grub1, grub2, grub1_efi, grub2_efi]))
        if len(_grubs) == 1:
            self.grub = _grubs[0]
            self.is_efi = self.grub._efi
        else:
            # Check if `/sys/firmware/efi` exist?
            self.is_efi = '/sys/firmware/efi' in sys_firmware if sys_firmware else False
            _grub1 = grub1_efi if self.is_efi else grub1
            _grub2 = grub2_efi if self.is_efi else grub2
            # Check grub version via installed-rpms
            if rpms:
                # grub1
                if 'grub2' not in rpms and 'grub' in rpms and _grub1 is not None:
                    self.grub = _grub1
                # grub2
                if 'grub' not in rpms and 'grub2' in rpms and _grub2 is not None:
                    self.grub = _grub2
            # Check grub version via the booted CmdLine
            if self.grub is None and cmdline:
                # grub1
                if "BOOT_IMAGE" not in cmdline or 'rd_LVM_LV' in cmdline:
                    self.grub = _grub1
                # grub2
                if "BOOT_IMAGE" in cmdline or 'rd.lvm.lv' in cmdline:
                    self.grub = _grub2

        if self.grub:
            self.version = self.grub._version
            self.is_kdump_iommu_enabled = self.grub.is_kdump_iommu_enabled
            self.kernel_initrds = self.grub.kernel_initrds
        else:
            raise ParseException('No valid grub configuration is found.')

    @defaults([])
    def get_grub_cmdlines(self, search_text=None):
        """
        Get the boot entries in which `cmdline` contains the `search_text`,
        return all the boot entries by default.

        Arguments:
            search_text(str): keyword to find in the `cmdline`, being set to
                None by default.

        Returns:
            A list of :class:`insights.parsers.grub_conf.BootEntry` objects for
            each boot entry in which the `cmdline` contains the `search_text`.
        """
        if search_text is None:
            return [entry for entry in self.grub.boot_entries]
        elif search_text:
            return [entry for entry in self.grub.boot_entries if search_text in entry.cmdline]
        return []



"""
Parsers for usage of VirtualDirContext option in Tomcat config files
====================================================================

This module provides the following parsers:

TomcatVirtualDirContextFallback
-------------------------------

This is a parser for a command for finding config files in default location:

``/usr/bin/find /usr/share -maxdepth 1 -name 'tomcat*' -exec grep -R -s 'VirtualDirContext' --include '*.xml' '{}' +``

It is especially useful if the Tomcat server is not running.


TomcatVirtualDirContextTargeted
-------------------------------

This is a parser for a command for finding config files in the custom location defined in a command
line:

``/bin/grep -R -s 'VirtualDirContext' --include '*.xml' {catalina}``

Where ``catalina`` variable is computed as following:

``/bin/ps auxww | awk '/java/ { match($0, "\\-Dcatalina\\.home=([^[:space:]]+)", a); match($0, "\\-Dcatalina\\.base=([^[:space:]]+)", b); if (a[1] != "" || b[1] != "") print a[1] " " b[1] }'``


Both parsers detect whether there are any config files which contain VirtualDirContext.

Sample input::

    /usr/share/tomcat/conf/server.xml:    <Resources className="org.apache.naming.resources.VirtualDirContext"

Examples::

    >>> shared[TomcatVirtualDirContextFallback].data
    {'/usr/share/tomcat/conf/server.xml':
     ['    <Resources className="org.apache.naming.resources.VirtualDirContext"'],
     }
"""

from .. import parser, CommandParser
from ..specs import Specs
from . import SkipException


class TomcatVirtualDirContextBase(CommandParser):
    """
    Generic parser which finds whether there is a VirtualDirContext option used in TomCat
    configuration file.
    """
    def __init__(self, *args, **kwargs):
        self.data = {}
        super(TomcatVirtualDirContextBase, self).__init__(*args, **kwargs)

    def parse_content(self, content):
        for line in content:
            try:
                file_name, file_line = line.split(':', 1)  # Hoping that : is not part of the path
            except ValueError:
                continue  # skip this line
            if file_name.endswith('.xml'):  # Make sure that nothing else is stored
                if file_name in self.data:
                    self.data[file_name].append(file_line)
                else:
                    self.data[file_name] = [file_line]

        if self.data == {}:
            raise SkipException('VirtualDirContext not used.')


@parser(Specs.tomcat_vdc_fallback)
class TomcatVirtualDirContextFallback(TomcatVirtualDirContextBase):
    """
    Reports whether there is a VirtualDirContext option used in TomCat configuration file. Looks
    for the configuration files in default location.
    """
    pass


@parser(Specs.tomcat_vdc_targeted)
class TomcatVirtualDirContextTargeted(TomcatVirtualDirContextBase):
    """
    Reports whether there is a VirtualDirContext option used in TomCat configuration file. Looks
    for the configuration files in location derived from running Tomcat command.
    """
    pass



"""
DnsmasqConf - files ``/etc/dnsmasq.conf`` and ``/etc/dnsmasq.d/*.conf``
=======================================================================
"""

from insights.core.plugins import parser
from insights.core import ConfigParser
from insights.specs import Specs
from insights.configtree import (
    DocParser,
    LineGetter,
    parse_name_attrs,
    Node,
)


class DnsmasqConfDocParser(DocParser):
    def parse_statement(self, lg):
        line = lg.peek()
        pos = lg.pos
        if "=" in line:
            name, attrs = parse_name_attrs(line, sep="=")
        else:
            name, attrs = line, None

        ret = Node(name=name, attrs=attrs, ctx=self.ctx)
        next(lg)
        ret.pos = pos
        return ret


def parse_doc(f, ctx=None):
    return DnsmasqConfDocParser(ctx).parse_doc(LineGetter(f))


@parser(Specs.dnsmasq_config)
class DnsmasqConf(ConfigParser):
    """
    Class to parses the content of dnsmasq configuration files ``/etc/dnsmasq.conf`` and ``/etc/dnsmasq.d/*.conf``

    Sample configuration output::

        # Listen on this specific port instead of the standard DNS port
        # (53). Setting this to zero completely disables DNS function,
        # leaving only DHCP and/or TFTP.
        port=5353

        no-resolv
        domain-needed
        no-negcache
        max-cache-ttl=1
        enable-dbus
        dns-forward-max=5000
        cache-size=5000
        bind-dynamic
        except-interface=lo
        server=/in-addr.arpa/127.0.0.1
        server=/cluster.local/127.0.0.1
        # End of config

    Examples::
        >>> "no-resolv" in conf
        True
        >>> conf["server"]
        server=/in-addr.arpa/127.0.0.1
        server=/cluster.local/127.0.0.1
        >>> conf["dns-forward-max"][-1].value
        5000

    """

    def parse_doc(self, content):
        return parse_doc(content, ctx=self)



"""
InitProcessCgroup - File ``/proc/1/cgroup``
===========================================

This parser reads the content of ``/proc/1/cgroup``.
This file shows the cgroup detail of init process.
The format of the content is like key-value. We can
also use this info to check if the archive is from
container or host.
"""

from .. import parser, CommandParser, LegacyItemAccess
from insights.specs import Specs


@parser(Specs.init_process_cgroup)
class InitProcessCgroup(CommandParser, LegacyItemAccess):
    """
    Class ``InitProcessCgroup`` parses the content of the ``/proc/1/cgroup``.

    Attributes:

        is_container (bool): It is used to check if a archive is from host or container.
            Return True if the archive is from container.

    A small sample of the content of this file looks like::

        11:hugetlb:/
        10:memory:/
        9:devices:/
        8:pids:/
        7:perf_event:/
        6:net_prio,net_cls:/
        5:blkio:/
        4:freezer:/
        3:cpuacct,cpu:/
        2:cpuset:/
        1:name=systemd:/

    Examples:
        >>> type(cgroupinfo)
        <class 'insights.parsers.init_process_cgroup.InitProcessCgroup'>
        >>> cgroupinfo["memory"]
        ["10", "/"]
        >>> cgroupinfo.is_container
        False
    """

    def parse_content(self, content):
        self.data = {}
        self.is_container = False
        for line in content:
            values = line.split(":")
            self.data[values[1]] = [values[0], values[2]]
            if "system.slice/docker-" in values[2] or 'machine.slice/libpod-' in values[2]:
                self.is_container = True



"""
IsRhel6, IsRhel7 and IsRhel8
===============================

The ``IsRhel*`` components each use the ``RedhatRelease`` combiner to
retrieve the RHEL version information.
Each component checks if the release version matches the version it represents,
if the version does not match what is expected the class raises ``SkipComponent``
so that the dependent component will not fire.
Can be added as a dependency of a parser so that the parser only fires if the
``IsRhel*`` dependency is met.

An example from the following ``/etc/redhat_release`` file output::

    Red Hat Enterprise Linux release 8.0 (Ootpa)

Example:

    >>> type(IsRhel8)
    <class 'insights.components.rhel_version.Is_Rhel8'>
    >>> is_rhel8.rhel_version
    '8.0'
"""

from insights.core.plugins import component
from insights.combiners.redhat_release import RedHatRelease
from insights.core.dr import SkipComponent


@component(RedHatRelease)
class IsRhel6(object):
    """
    This component uses ``RedHatRelease`` combiner
    to determine RHEL version. It checks if RHEL6, if not
    RHEL6 it raises ``SkipComponent``.

    Raises:
        SkipComponent: When RHEL version is not RHEL6.
    """
    def __init__(self, rhel):
        if rhel.major != 6:
            raise SkipComponent('Not RHEL6')


@component(RedHatRelease)
class IsRhel7(object):
    """
    This component uses ``RedHatRelease`` combiner
    to determine RHEL version. It checks if RHEL7, if not \
    RHEL7 it raises ``SkipComponent``.

    Raises:
        SkipComponent: When RHEL version is not RHEL7.
    """
    def __init__(self, rhel):
        if rhel.major != 7:
            raise SkipComponent('Not RHEL7')


@component(RedHatRelease)
class IsRhel8(object):
    """
    This component uses ``RedhatRelease`` combiner
    to determine RHEL version. It checks if RHEL8, if not
    RHEL8 it raises ``SkipComponent``.

    Raises:
        SkipComponent: When RHEL version is not RHEL8.
    """
    def __init__(self, rhel):
        if rhel.major != 8:
            raise SkipComponent('Not RHEL8')



"""
The filters module allows developers to apply filters to datasources. A filter
is a simple string, and it matches if it is contained anywhere within a line.

If a datasource has filters defined, it will return only lines matching at
least one of them. If a datasource has no filters, it will return all lines.

Filters aren't applicable to "raw" datasources, which are created with
``kind=RawFileProvider`` and have ``RegistryPoint`` instances with ``raw=True``.

The addition of a single filter can cause a datasource to change from returning
all lines to returning just those that match. Therefore, any filtered
datasource should have at least one filter in the commit introducing it so
downstream components don't inadvertently change its behavior.

The benefit of this fragility is the ability to drastically reduce in-memory
footprint and archive sizes. An additional benefit is the ability to evaluate
only lines known to be free of sensitive information.

Filters added to a ``RegistryPoint`` will be applied to all datasources that
implement it. Filters added to a datasource implementation apply only to that
implementation.

For example, a filter added to ``Specs.ps_auxww`` will apply to
``DefaultSpecs.ps_auxww``, ``InsightsArchiveSpecs.ps_auxww``,
``SosSpecs.ps_auxww``, etc. But a filter added to `DefaultSpecs.ps_auxww` will
only apply to ``DefaultSpecs.ps_auxww``. See the modules in ``insights.specs``
for those classes.

Filtering can be disabled globally by setting the environment variable
``INSIGHTS_FILTERS_ENABLED=False``. This means that no datasources will be
filtered even if filters are defined for them.
"""
import os
import pkgutil
import six
import yaml as ser
from collections import defaultdict

import insights
from insights.core import dr, plugins
from insights.util import parse_bool

_CACHE = {}
FILTERS = defaultdict(set)
ENABLED = parse_bool(os.environ.get("INSIGHTS_FILTERS_ENABLED"), default=True)


def add_filter(ds, patterns):
    """
    Add a filter or list of filters to a datasource. A filter is a simple
    string, and it matches if it is contained anywhere within a line.

    Args:
       ds (@datasource component): The datasource to filter
       patterns (str, [str]): A string, list of strings, or set of strings to
            add to the datasource's filters.
    """
    if not plugins.is_datasource(ds):
        raise Exception("Filters are applicable only to datasources.")

    delegate = dr.get_delegate(ds)

    if delegate.raw:
        raise Exception("Filters aren't applicable to raw datasources.")

    if not delegate.filterable:
        raise Exception("Filters aren't applicable to %s." % dr.get_name(ds))

    if ds in _CACHE:
        del _CACHE[ds]
    if isinstance(patterns, six.string_types):
        FILTERS[ds].add(patterns)
    elif isinstance(patterns, list):
        FILTERS[ds] |= set(patterns)
    elif isinstance(patterns, set):
        FILTERS[ds] |= patterns
    else:
        raise TypeError("patterns must be string, list, or set.")


def get_filters(component):
    """
    Get the set of filters for the given datasource.

    Filters added to a ``RegistryPoint`` will be applied to all datasources that
    implement it. Filters added to a datasource implementation apply only to
    that implementation.

    For example, a filter added to ``Specs.ps_auxww`` will apply to
    ``DefaultSpecs.ps_auxww``, ``InsightsArchiveSpecs.ps_auxww``,
    ``SosSpecs.ps_auxww``, etc. But a filter added to ``DefaultSpecs.ps_auxww``
    will only apply to ``DefaultSpecs.ps_auxww``. See the modules in
    ``insights.specs`` for those classes.

    Args:
        component (a datasource): The target datasource

    Returns:
        set: The set of filters defined for the datasource
    """
    def inner(c, filters=None):
        filters = filters or set()
        if not ENABLED:
            return filters

        if not plugins.is_datasource(c):
            return filters

        if c in FILTERS:
            filters |= FILTERS[c]

        for d in dr.get_dependents(c):
            filters |= inner(d, filters)
        return filters

    if component not in _CACHE:
        _CACHE[component] = inner(component)
    return _CACHE[component]


def apply_filters(target, lines):
    """
    Applys filters to the lines of a datasource. This function is used only in
    integration tests. Filters are applied in an equivalent but more performant
    way at run time.
    """
    filters = get_filters(target)
    if filters:
        for l in lines:
            if any(f in l for f in filters):
                yield l
    else:
        for l in lines:
            yield l


_filename = ".".join(["filters", ser.__name__])
_dumps = ser.dump
_loads = ser.safe_load


def loads(string):
    """Loads the filters dictionary given a string."""
    d = _loads(string)
    for k, v in d.items():
        FILTERS[dr.get_component(k) or k] = set(v)


def load(stream=None):
    """
    Loads filters from a stream, normally an open file. If one is
    not passed, filters are loaded from a default location within
    the project.
    """
    if stream:
        loads(stream.read())
    else:
        data = pkgutil.get_data(insights.__name__, _filename)
        return loads(data) if data else None


def dumps():
    """Returns a string representation of the FILTERS dictionary."""
    d = {}
    for k, v in FILTERS.items():
        d[dr.get_name(k)] = list(v)
    return _dumps(d)


def dump(stream=None):
    """
    Dumps a string representation of `FILTERS` to a stream, normally an
    open file. If none is passed, `FILTERS` is dumped to a default location
    within the project.
    """
    if stream:
        stream.write(dumps())
    else:
        path = os.path.join(os.path.dirname(insights.__file__), _filename)
        with open(path, "wu") as f:
            f.write(dumps())



"""
TeamdctlStateDump - command ``teamdctl {team interface} state dump``
====================================================================

This module provides processing for the output of the command
``teamdctl {team interface} state dump`` which is JSON pattern.

Examples:

    >>> teamdctl_state_dump_content = '''
    ... {
    ...     "runner": {
    ...         "active_port": "eno1"
    ...     },
    ...     "setup": {
    ...         "daemonized": false,
    ...         "dbus_enabled": true,
    ...         "debug_level": 0,
    ...         "kernel_team_mode_name": "activebackup",
    ...         "pid": 4464,
    ...         "pid_file": "/var/run/teamd/team0.pid",
    ...         "runner_name": "activebackup",
    ...         "zmq_enabled": false
    ...     },
    ...     "team_device": {
    ...         "ifinfo": {
    ...             "dev_addr": "2c:59:e5:47:a9:04",
    ...             "dev_addr_len": 6,
    ...             "ifindex": 5,
    ...             "ifname": "team0"
    ...         }
    ...     }
    ... }
    ... '''.strip()

    >>> from insights.parsers.teamdctl_state_dump import TeamdctlStateDump
    >>> from insights.tests import context_wrap
    >>> shared = {TeamdctlStateDump: TeamdctlStateDump(context_wrap(teamdctl_state_dump_content))}
    >>> result = shared[TeamdctlStateDump]
    >>> result['runner']['active_port']
    'eno1'
    >>> result.runner_type
    'activebackup'
    >>> result.team_ifname
    'team0'
"""

from .. import JSONParser, parser, defaults, CommandParser
from insights.specs import Specs


@parser(Specs.teamdctl_state_dump)
class TeamdctlStateDump(CommandParser, JSONParser):
    """
    Class to parse the output of ``teamdctl {team interface} state dump``.
    """
    @property
    @defaults()
    def runner_type(self):
        """
        str: Return the type of the teaming runner
        """
        return self.data["setup"]["runner_name"]

    @property
    @defaults()
    def team_ifname(self):
        """
        str: Return the teaming device name
        """
        return self.data["team_device"]["ifinfo"]["ifname"]



import doctest

from .test_rabbitmq_queues import QUEUES
from insights.parsers import rabbitmq
from insights.tests import context_wrap


RABBITMQ_ENV = """
RABBITMQ_SERVER_ERL_ARGS="+K true +P 1048576 -kernel inet_default_connect_options [{nodelay,true},{raw,6,18,<<5000:64/native>>}] -kernel inet_default_listen_options [{raw,6,18,<<5000:64/native>>}]"
""".strip()

RABBITMQ_ENV_DIFFERENT_TIMEOUT = """
RABBITMQ_SERVER_ERL_ARGS="+K true +P 1048576 -kernel inet_default_connect_options [{nodelay,true},{raw,6,18,<<5000:64/native>>}] -kernel inet_default_listen_options [{raw,6,18,<<3000:64/native>>}]"
""".strip()

RABBITMQ_ENV_OPTIONS = """
OPTIONS="+K true +P 1048576 -kernel inet_default_connect_options [{nodelay,true},{raw,6,18,<<5000:64/native>>}] -kernel inet_default_listen_options [{raw,6,18,<<5000:64/native>>}]"
""".strip()

RABBITMQ_ENV_BAD_PATTERN = """
RABBITMQ_SERVER_ERL_ARGS="+K true +P 1048576 -kernel inet_default_connect_options [{nodelay,true},{raw,6,18,<<5000:64/native>>}] -kernel inet_default_options [{raw,6,18,<<5000:64/native>>}]"
""".strip()


def test_rabbitmq_env():
    rabbitmq_env = rabbitmq.RabbitMQEnv(context_wrap(RABBITMQ_ENV))
    assert rabbitmq_env.rabbitmq_server_erl_args == '+K true +P 1048576 -kernel inet_default_connect_options [{nodelay,true},{raw,6,18,<<5000:64/native>>}] -kernel inet_default_listen_options [{raw,6,18,<<5000:64/native>>}]'
    assert rabbitmq_env.data['RABBITMQ_SERVER_ERL_ARGS'] == '+K true +P 1048576 -kernel inet_default_connect_options [{nodelay,true},{raw,6,18,<<5000:64/native>>}] -kernel inet_default_listen_options [{raw,6,18,<<5000:64/native>>}]'
    assert rabbitmq_env.rmq_erl_tcp_timeout == '5000'

    rabbitmq_env = rabbitmq.RabbitMQEnv(context_wrap(RABBITMQ_ENV_DIFFERENT_TIMEOUT))
    assert rabbitmq_env.rmq_erl_tcp_timeout is None

    rabbitmq_env = rabbitmq.RabbitMQEnv(context_wrap(RABBITMQ_ENV_OPTIONS))
    assert rabbitmq_env.rabbitmq_server_erl_args is None
    assert rabbitmq_env.rmq_erl_tcp_timeout is None

    rabbitmq_env = rabbitmq.RabbitMQEnv(context_wrap(RABBITMQ_ENV_BAD_PATTERN))
    assert rabbitmq_env.rmq_erl_tcp_timeout is None


def test_doc_examples():
    failed, total = doctest.testmod(
        rabbitmq,
        globs={'rabbitmq_env': rabbitmq.RabbitMQEnv(context_wrap(RABBITMQ_ENV)),
               'queues': rabbitmq.RabbitMQQueues(context_wrap(QUEUES))}
    )
    assert failed == 0



"""
dmsetup commands - Command ``dmsetup``
======================================

Parsers for parsing and extracting data from output of commands related to
``dmsetup``.
Parsers contained in this module are:

DmsetupInfo - command ``dmsetup info -C``
-----------------------------------------

"""

from insights import parser, CommandParser
from insights.parsers import parse_delimited_table
from insights.specs import Specs


@parser(Specs.dmsetup_info)
class DmsetupInfo(CommandParser):
    """
    ``dmsetup info -C`` command output

    Example input::

        Name               Maj Min Stat Open Targ Event  UUID
        VG00-tmp           253   8 L--w    1    1      0 LVM-gy9uAwD7LuTIApplr2sogbOx5iS0FTax6lLmBji2ueSbX49gxcV76M29cmukQiw4
        VG00-home          253   3 L--w    1    1      0 LVM-gy9uAwD7LuTIApplr2sogbOx5iS0FTaxCqXOnbGe2zjhX923dFiIdl1oi7mO9tXp
        VG00-var           253   6 L--w    1    2      0 LVM-gy9uAwD7LuTIApplr2sogbOx5iS0FTaxicvyvt67113nTb8vMlGfgdEjDx0LKT2O
        VG00-swap          253   1 L--w    2    1      0 LVM-gy9uAwD7LuTIApplr2sogbOx5iS0FTax3Ll2XhOYZkylx1CjOQi7G4yHgrIOsyqG
        VG00-root          253   0 L--w    1    1      0 LVM-gy9uAwD7LuTIApplr2sogbOx5iS0FTaxKpnAKYhrYMYMNMwjegkW965bUgtJFTRY
        VG00-var_log_audit 253   5 L--w    1    1      0 LVM-gy9uAwD7LuTIApplr2sogbOx5iS0FTaxwQ8R0XWJRm86QX3befq1cHRy47Von6ZW

    Example data structure produced::

        data = [
          {
            'Stat': 'L--w',
            'Name': 'VG00-tmp',
            'Min': '8',
            'Targ': '1',
            'Maj': '253',
            'Open': '1',
            'Event': '0',
            'UUID': 'LVM-gy9uAwD7LuTIApplr2sogbOx5iS0FTax6lLmBji2ueSbX49gxcV76M29cmukQiw4'
          },...
        ]

    Attributes:
        data (list): List of devices found, in order
        names (list): Device names, in order found
        uuids (list): UUID
        by_name (dict): Access to each device by devicename
        by_uuid (dict): Access to each device by uuid

    Example:
        >>> len(info)
        6
        >>> info.names[0]
        'VG00-tmp'
        >>> info[1]['Maj']
        '253'
        >>> info[1]['Stat']
        'L--w'
    """

    def parse_content(self, content):
        self.data = parse_delimited_table(content)

        self.names = [dm['Name'] for dm in self.data if 'Name' in dm]
        self.by_name = dict((dm['Name'], dm) for dm in self.data if 'Name' in dm)
        self.uuids = [dm['UUID'] for dm in self.data if 'UUID' in dm]
        self.by_uuid = dict((dm['UUID'], dm) for dm in self.data if 'UUID' in dm)

    def __len__(self):
        """
        The length of the devices list
        """
        return len(self.data)

    def __iter__(self):
        """
        Iterate through the devices list
        """
        for dm in self.data:
            yield dm

    def __getitem__(self, idx):
        """
        Fetch a device by index in devices list
        """
        return self.data[idx]



import json
import atexit
import hashlib
import os
import re
import shutil
import subprocess
import shlex
import tempfile
from contextlib import contextmanager

from insights.config import CommandSpec
from insights.config.static import get_config
from insights.parsers.uname import rhel_release_map
from insights.core import context as ctx
from insights.util import fs

_data_spec_config = get_config()

REPO_PATH = "./repository"


class TempMaker(object):

    def __init__(self, cleanup=True, path='/tmp'):
        self.cleanup = cleanup
        self.path = tempfile.mkdtemp(suffix='_insights_archive', dir=path)
        self.temps = set()

    def get_temp(self):
        path = tempfile.mkdtemp(dir=self.path)
        self.temps.add(path)
        return path

    def close(self):
        if self.cleanup:
            fs.remove(self.path)


__tmp_maker = None


def __cleanup():
    if __tmp_maker:
        __tmp_maker.close()


def get_temp_dir():
    global __tmp_maker
    if not __tmp_maker:
        __tmp_maker = TempMaker()
        atexit.register(__cleanup)
    return __tmp_maker.get_temp()


class BaseArchive(object):

    root_prefix = "base_archives"

    def __init__(self, name, validate=True):
        self.name = name
        root_prefix = os.path.abspath(os.path.join(os.path.dirname(os.path.abspath(__file__)),
                                      REPO_PATH,
                                      self.root_prefix))
        if not os.path.exists(root_prefix):
            os.makedirs(root_prefix)
        self.root_dir = os.path.join(root_prefix, name)
        if validate:
            self.validate()

    def validate(self):
        if not self.name:
            raise ValueError("Name not valid")
        if not os.path.exists(self.root_dir):
            raise IOError("{0} {1} can not be found".format(self.__class__.__name__, self.name))

    def files(self, absolute=False):
        prefix_len = 0 if absolute else len(self.root_dir) + 1
        for dirpath, _, filenames in os.walk(self.root_dir):
            for filename in filenames:
                yield os.path.join(dirpath[prefix_len:], filename)


def strip_slash(path):
    path = path.strip()
    if path.startswith("/"):
        path = path.lstrip("/")
    return path


class TempOverlay(BaseArchive):

    def __init__(self, files):
        self.name = "TempOverlay"
        self.cached_files = files if files else []
        self.root_dir = get_temp_dir()
        for i, t in enumerate(self.cached_files):
            path, content = t
            path = strip_slash(path)
            self.cached_files[i] = (path, content)
            self.add_file(path, content)

    def add_file(self, path, content):
        if not isinstance(path, str):
            raise ValueError("Invalid path type: {0}".format(type(path)))
        # We don't allow absolute paths
        if content:
            full_path = os.path.join(self.root_dir, path)
            if not os.path.exists(os.path.dirname(full_path)):
                os.makedirs(os.path.dirname(full_path))
            with open(full_path, "w") as f:
                f.write(content.encode("utf-8") + "\n")

    def files(self, absolute=False):
        for path, content in self.cached_files:
            if content:
                yield os.path.join(self.root_dir, path) if absolute else path


class Overlay(BaseArchive):

    root_prefix = "overlays"


class Transform(object):

    def __init__(self, path, force_create=True):
        self.path = path
        self.ops = []
        self.paths = []
        self.force_create = force_create

    def resolve_paths(self, archive):
        specs = _data_spec_config.get_spec_list(self.path)
        if not specs:
            specs = _data_spec_config.get_meta_spec_list(self.path)
            if not specs:
                raise ValueError("Invalid symbolic name: [%s]" % self.path)
        for archive_file in archive.files():
            for spec in specs:
                if spec.matches(archive_file, suffix="$"):
                    self.paths.append(archive_file)
        if not self.paths:
            if self.force_create:
                primary_value = specs[0]
                if isinstance(primary_value, CommandSpec):
                    new_path = os.path.join(archive.root_dir, "insights_commands", primary_value.get_path())
                else:
                    new_path = os.path.join(archive.root_dir, primary_value.get_path())
                self.paths.append(new_path)
            else:
                raise ValueError("Symbolic name not found: {0}".format(self.path))

    def replace(self, content):
        def _replace(path):
            parent_dir = os.path.dirname(path)
            if not os.path.exists(parent_dir):
                os.makedirs(parent_dir)
            with open(path, 'w') as fp:
                fp.write(content.encode("utf-8") + "\n")
        self.ops.append(_replace)
        return self

    def append(self, content):
        def _append(path):
            with open(path, 'a') as fp:
                fp.write(content.encode("utf-8") + "\n")
        self.ops.append(_append)
        return self

    def sub(self, pattern, replacement):
        pattern = make_regexp(pattern)

        def _sub(path):
            with file_replacer(path) as (fp, tf):
                for line in fp:
                    tf.write(pattern.sub(replacement, line))

        self.ops.append(_sub)
        return self

    def grep(self, pattern, keep=True):
        pattern = make_regexp(pattern)

        def _grep(path):
            with file_replacer(path) as (fp, tf):
                for line in fp:
                    if bool(pattern.match(line)) == keep:  # xor
                        tf.write(line)
                    else:
                        continue

        self.ops.append(_grep)
        return self

    def exclude(self, pattern):
        return self.grep(pattern, keep=False)

    def execute(self, root_dir):
        for op in self.ops:
            for path in self.paths:
                op(os.path.join(root_dir, path))
        self.paths = []


class TestArchive(BaseArchive):

    root_prefix = "test_archives"

    def __init__(self, name, base_archive="rhel7", overlays=None, transforms=None, machine_id=None, removals=None, compression="gz", hostname=None, **kwargs):
        super(TestArchive, self).__init__(name, validate=False)
        if not os.path.exists(self.root_dir):
            os.mkdir(self.root_dir)
        self.base_archive = (
            base_archive if isinstance(base_archive, BaseArchive) else BaseArchive(base_archive))
        self.overlays = [Overlay(o) for o in overlays] if overlays else []
        self.transforms = transforms if transforms else []
        self.machine_id = machine_id if machine_id else self.generate_machine_id()
        self.machine_id_override = bool(machine_id)
        self.removals = removals if removals else []
        self.compression = compression
        self.hostname = hostname if hostname else self.machine_id
        self.extra = kwargs

    def create_dir_structure(self):
        self.copy_archive(self.base_archive)
        for spec in self.removals:
            if isinstance(spec, str):
                spec = _data_spec_config.get_spec_list(spec)[0]
            for f in filter(spec.matches, self.files()):
                path_to_unlink = os.path.join(self.root_dir, f)
                if os.path.exists(path_to_unlink):
                    os.unlink(path_to_unlink)

        for o in self.overlays:
            self.copy_archive(o)
        for t in self.transforms:
            t.resolve_paths(self)
            t.execute(self.root_dir)
        self.apply_metadata()

    def apply_plugin(self, plugin):
        self.transform(*plugin.get("transforms", []))
        self.overlay(*plugin.get("overlays", []))

    def apply_metadata(self):
        machine_id_path = os.path.join(self.root_dir, "etc/redhat-access-insights")

        if not os.path.exists(machine_id_path):
            os.makedirs(machine_id_path)

        with open(os.path.join(machine_id_path, "machine-id"), "w") as f:
            f.write(self.machine_id if self.machine_id_override else self.generate_machine_id())
        with open(os.path.join(self.root_dir, "branch_info"), "w") as f:
            f.write(self.generate_branch_info())

    @contextmanager
    def file_content(self, name):
        with open(os.path.join(self.root_dir, name), "r") as fd:
            yield fd

    def transform(self, *t):
        for transform in t:
            self.transforms.append(transform)
        return self

    def overlay(self, *o):
        for overlay in o:
            self.overlays.append(overlay)
        return self

    def get_context(self):
        return ctx.Context(hostname=self.hostname,
                           machine_id=self.machine_id,
                           **self.extra)

    def generate_machine_id(self):
        h = hashlib.sha224(self.name)
        h.update(self.base_archive.name)
        for overlay in self.overlays:
            h.update(overlay.name)
        for transform in self.transforms:
            h.update(transform.path)
        return "TEST-" + h.hexdigest()

    def generate_branch_info(self):
        return '{"remote_branch": -1, "remote_leaf": -1}'

    def copy_archive(self, archive):
        for fname in archive.files():
            src = os.path.join(archive.root_dir, fname)
            dst = os.path.join(self.root_dir, fname)
            dstdir = os.path.dirname(dst)
            if not os.path.exists(dstdir):
                os.makedirs(dstdir)
            shutil.copyfile(src, dst)

    def export_options(self):
        if self.compression == "gz":
            return "czf", "tar.gz"
        elif self.compression == "xz":
            return "cJf", "tar.xz"
        else:
            return "cf", "tar"

    def export(self, dest=".", nested_root=True):
        options, ext = self.export_options()
        if nested_root:
            dir_root = os.path.join(self.root_dir, "..")
            root_name = self.name
        else:
            dir_root = self.root_dir
            root_name = "."
        subprocess.check_call(
            shlex.split("tar {0} {1}.{2} -C {3} {4}".format(
                options, os.path.join(dest, self.name),
                ext, dir_root, root_name)))
        return self.output_path(dest)

    def output_path(self, dest):
        return os.path.join(dest, "{0}.{1}".format(
            self.name,
            self.export_options()[1]))

    def clean(self):
        fs.remove(self.root_dir)

    def build(self, dest, force=False):
        if force:
            self.clean()
        elif os.path.exists(self.output_path(dest)):
            return self.output_path(dest)
        self.create_dir_structure()
        return self.export(dest)

    def apply_input_data(self, input_data):
        self._apply_input_data_removals(input_data)
        self._apply_input_data_content(input_data)
        if input_data.hostname != "hostname.example.com":
            self.transforms.append(Transform("hostname").replace(input_data.hostname))
        self.machine_id = input_data.machine_id
        self.machine_id_override = True
        if input_data.release != "default-release":
            self.transforms.append(Transform("redhat-release").replace(input_data.release))
        if (input_data.version != ["-1", "-1"] and
                not [t.path for t in self.transforms if t.path == "uname"]):
            rhel_version = ".".join(input_data.version)
            for kernel, rhel in rhel_release_map.items():
                if rhel_version == rhel:
                    nvr_regex = " \d*\.\d*\.\d*-\d*"
                    self.transforms.append(Transform("uname").sub(nvr_regex, " " + kernel))

    def _apply_input_data_removals(self, input_data):
        for symbolic_name in input_data.to_remove:
            self.removals.extend(_data_spec_config.get_spec_list(symbolic_name))

    def _apply_input_data_content(self, input_data):
        for record in input_data.records:
            target = record["target"]
            content = record["context"].content
            path = record["context"].path
            if path and "BOGUS" not in path:
                # Presence of path means it's a pattern file, which means it
                # probably won't exist in the current archive
                self.overlays.append(TempOverlay([(path, content)]))
            else:
                self.transforms.append(Transform(target).replace(content))


class MultiArchive(TestArchive):

    def __init__(self, name, archives=None, machine_id=None, display_name=None, **kwargs):
        super(MultiArchive, self).__init__(name, machine_id=machine_id, **kwargs)
        self.archives = archives if archives else []
        self.metadata = None
        self.display_name = display_name
        self.extra_metadata = {}

    def copy_archive(self, archive):
        pass

    def apply_metadata(self):
        pass

    def add_metadata(self, md):
        self.extra_metadata.update(md)
        return self

    def build(self, dest, force=True):
        if force:
            self.clean()
        elif os.path.exists(self.output_path(dest)):
            return self.output_path(dest)
        if not self.metadata:
            self.metadata = self.build_metadata(self.archives)
        self.transforms.append(self.metadata)
        self.create_dir_structure()
        for sub_archive in self.archives:
            sub_archive.compression = None
            sub_archive_path = sub_archive.build(dest, force=force)
            shutil.move(sub_archive_path, self.root_dir)
        return self.export(dest, nested_root=False)

    def build_metadata(self, sub_archives):
        first_ctx = sub_archives[0].get_context()
        product = first_ctx.product()
        parent = get_parent(sub_archives)
        systems = []
        for system in sub_archives:
            sys_ctx = system.get_context()
            systems.append({
                "product": product.name,
                "display_name": sys_ctx.hostname,
                "system_id": sys_ctx.machine_id,
                "type": sys_ctx.product().role,
                "links": build_links(system, sub_archives, parent)
            })
        metadata = {
            "product": product.name,
            "display_name": self.display_name if self.display_name else parent.hostname,
            "rhel_version": first_ctx.release,
            "system_id": self.machine_id if self.machine_id else first_ctx.machine_id,
            "systems": systems
        }
        metadata.update(self.extra_metadata)
        return Transform("metadata.json").replace(json.dumps(metadata))

    def apply_input_data(self, sub_archives):
        self.metadata = self.build_metadata(sub_archives)
        for i, input_data in enumerate(sub_archives):
            sub_archive = TestArchive("sub-archive-%d" % i, compression=None)
            sub_archive.apply_input_data(input_data)
            self.archives.append(sub_archive)


def build_links(target_archive, all_archives, parent):
    links = []
    is_parent = target_archive == parent
    if is_parent:
        for archive in all_archives:
            ctx = archive.get_context()
            links.append({
                "system_id": ctx.machine_id,
                "type": ctx.product().role
            })
    else:
        ctx = parent.get_context()
        links.append({
            "system_id": ctx.machine_id,
            "type": ctx.product().role
        })
    return links


def get_parent(sub_archives):
    for a in sub_archives:
        if a.get_context().product().is_parent():
            return a


@contextmanager
def file_replacer(path):
    with open(path, 'r') as fp:
        with tempfile.NamedTemporaryFile(delete=False) as tf:
            yield fp, tf
            tf.flush()
            temp_name = tf.name
    shutil.move(temp_name, path)


def make_regexp(pattern):
    if not hasattr(pattern, 'match'):
        pattern = re.compile(pattern)
    return pattern



"""
Sap - Combiner
==============

This combiner gets the running SAP instances on the system based on below
logic::

    if (SAPLOCALHOST = 'hostname') && InstanceType = D## ) then
        on this system runs SAP Netweaver Application Server version

    if (SAPLOCALHOST = 'hostname') && InstanceType = ASCS## ) then
        on this system runs SAP Netweaver Application Server Central Instance
        version

    if (SAPLOCALHOST = 'hostname') && InstanceType = HDB## ) then
        on this system runs SAP HANA database version

Check settings according SAP Notes compiled here:
https://wiki.scn.sap.com/wiki/x/rDK7Gg

"""

from collections import namedtuple
from insights import LegacyItemAccess
from insights.parsers import SkipException
from insights.core.plugins import combiner
from insights.combiners.hostname import hostname
from insights.parsers.lssap import Lssap
from insights.parsers.saphostctrl import SAPHostCtrlInstances


SAPInstances = namedtuple("SAPInstances",
        field_names=["name", "hostname", "sid", "type", "number", "version"])
"""namedtuple: Type for storing the SAP instance."""


@combiner(hostname, optional=[SAPHostCtrlInstances, Lssap])
class Sap(LegacyItemAccess):
    """
    Combiner for analyzing the SAP instances running on the system.

    Prefer SAPHostCtrlInstances to Lssap.

    Examples:
        >>> type(saps)
        <class 'insights.combiners.sap.Sap'>
        >>> saps['D16'].number
        '16'
        >>> saps.sid('HDB16')
        'HA2'
        >>> saps.hostname('HDB16')
        'lu0417'
        >>> 'D22' in saps.local_instances
        False
        >>> len(saps.business_instances)
        2
        >>> saps.is_hana
        True
        >>> saps.is_netweaver
        True
        >>> saps.is_ascs
        False

    Attributes:
        all_instances (list): List of all the SAP instances listed by the command.
        local_instances (list): List of SAP instances which are running on the system.
        function_instances (list): List of function SAP instances running on the system.
                                   E.g. Diagnostics Agents SMDA97/SMDA98
        business_instances (list): List of business SAP instances running on the system.
                                   E.g. HANA, NetWeaver, ASCS, or others
    """
    def __init__(self, hostname, insts, lssap):
        hn = hostname.hostname
        self.data = {}
        self.local_instances = []
        self.business_instances = []
        self.function_instances = []
        self.all_instances = []
        self._types = set()
        if insts:
            for inst in insts.data:
                k = inst['InstanceName']
                self.all_instances.append(k)
                if hn == inst['Hostname']:
                    self.local_instances.append(k)
                    self._types.add(inst['InstanceType'])
                self.data[k] = SAPInstances(k,
                                            inst['Hostname'],
                                            inst['SID'],
                                            inst['InstanceType'],
                                            inst['SystemNumber'],
                                            inst['SapVersionInfo'])
        elif lssap:
            for inst in lssap.data:
                k = inst['Instance']
                t = k.rstrip('1234567890')
                self.all_instances.append(k)
                if hn == inst['SAPLOCALHOST']:
                    self.local_instances.append(k)
                    self._types.add(t)
                self.data[k] = SAPInstances(k,
                                            inst['SAPLOCALHOST'],
                                            inst['SID'],
                                            t,
                                            inst['Nr'],
                                            inst['Version'])
        else:
            raise SkipException('No SAP instance.')

        FUNC_INSTS = ('SMDA')
        for i in self.local_instances:
            (self.function_instances if i.startswith(FUNC_INSTS) else self.business_instances).append(i)

    def version(self, instance):
        """str: Returns the version of the ``instance``."""
        return self.data[instance].version if instance in self.data else None

    def sid(self, instance):
        """str: Returns the sid of the ``instance``."""
        return self.data[instance].sid if instance in self.data else None

    def type(self, instance):
        """str: Returns the type code of the ``instance``."""
        return self.data[instance].type if instance in self.data else None

    def hostname(self, instance):
        """str: Returns the hostname of the ``instance``."""
        return self.data[instance].hostname if instance in self.data else None

    def number(self, instance):
        """str: Returns the systeme number of the ``instance``."""
        return self.data[instance].number if instance in self.data else None

    @property
    def is_netweaver(self):
        """bool: SAP NetWeaver is running on the system."""
        return 'D' in self._types

    @property
    def is_hana(self):
        """bool: SAP Hana is running on the system."""
        return 'HDB' in self._types

    @property
    def is_ascs(self):
        """bool: SAP System Central Services is running on the system."""
        return 'ASCS' in self._types



"""
VirtWhat
========

Combiner to check if the host is running on a virtual or physical machine.  It
uses the results of the ``DMIDecode`` and ``VirtWhat`` parsers.  Prefer
``VirtWhat`` to ``DMIDecode``.

Examples:
    >>> vw = shared[VirtWhat]
    >>> vw.is_virtual
    True
    >>> vw.is_physical
    False
    >>> vw.generic
    'kvm'
    >>> vw.amended_generic
    'rhev'
    >>> 'aws' in vw
    False
"""

from insights.core.plugins import combiner
from insights.parsers.dmidecode import DMIDecode
from insights.parsers.virt_what import VirtWhat as VW, BAREMETAL

# Below 2 Maps are only For DMIDecode
GENERIC_MAP = {
    'vmware': ['VMware'],
    'kvm': ['Red Hat', 'KVM'],
    'xen': ['Xen', 'domU'],
    'virtualpc': ['Microsoft Corporation', 'Virtual Machine'],
    'virtualbox': ['innotek GmbH'],
    'parallels': ['Parallels'],
    'qemu': ['QEMU'],
}

SPECIFIC_MAP = {
    'aws': ['amazon'],
    'xen-hvm': ['HVM'],
}


OVIRT = 'ovirt'
RHEV = 'rhev'


@combiner([DMIDecode, VW])
class VirtWhat(object):
    """
    A combiner for checking if this machine is virtual or physical by checking
    ``virt-what`` or ``dmidecode`` command.

    Prefer ``virt-what`` to ``dmidecode``

    Attributes:
        is_virtual (bool): It's running in a virtual machine?
        is_physical (bool): It's running in a physical machine?
        generic (str): The type of the virtual machine. 'baremetal' if physical machine.
        specifics (list): List of the specific information.
        amended_generic (str):The type of the virtual machine. 'baremetal' if physical machine.
            Added to address an issue with virt_what/dmidecode when identifying 'ovirt' vs 'rhev'.
            Will match the generic attribute in all other cases.
    """

    def __init__(self, dmi, vw):
        self.is_virtual = self.is_physical = None
        self.generic = ''
        self.specifics = []

        if vw and not vw.errors:
            self.is_physical = vw.is_physical
            self.is_virtual = vw.is_virtual
            self.generic = vw.generic
            self.specifics = vw.specifics

        # Error occurred in ``virt-what``, try ``dmidecode``
        if (vw is None or vw.errors) and dmi:
            sys_info = dmi.get("system_information", [{}])[0]
            bios_info = dmi.get("bios_information", [{}])[0]
            dmi_info = list(sys_info.values()) + list(bios_info.values())
            if dmi_info:
                for dmi_v in dmi_info:
                    if not self.generic:
                        generic = [g for g, val in GENERIC_MAP.items() if any(v in dmi_v for v in val)]
                        self.generic = generic[0] if generic else ''
                    self.specifics.extend([g for g, val in SPECIFIC_MAP.items() if any(v in dmi_v for v in val)])

                self.is_virtual = True
                self.is_physical = False
                if self.generic == '':
                    self.generic = BAREMETAL
                    self.is_virtual = False
                    self.is_physical = True

        sys_info = dmi.get("system_information", [{}])[0] if dmi else None

        self.amended_generic = (RHEV if sys_info['product_name'].lower() == 'rhev hypervisor' else
                                OVIRT if sys_info['product_name'].lower() == 'ovirt node' else
                                 self.generic) if sys_info else self.generic

    def __contains__(self, name):
        """bool: Is this ``name`` found in the specific list?"""
        return name in [self.generic] + self.specifics



import pytest

from insights.core import YAMLParser, ParseException
from insights.tests import context_wrap


bi_conf_content = """
{"remote_branch": -1, "remote_leaf": -1}
""".strip()


class FakeYamlParser(YAMLParser):
    """ Class for parsing the content of ``branch_info``."""
    pass


class MyYamlParser(YAMLParser):
    pass


def test_json_parser_success():
    ctx = context_wrap('a: 2')
    assert FakeYamlParser(ctx)


def test_json_parser_failure():
    ctx = context_wrap("boom /")
    with pytest.raises(ParseException) as ex:
        FakeYamlParser(ctx)

    assert "FakeYamlParser" in ex.value.args[0]


def test_settings_yml():
    ctx = context_wrap(bi_conf_content)
    ctx.content = bi_conf_content
    result = FakeYamlParser(ctx)
    assert result.data['remote_branch'] == -1
    assert result.data['remote_leaf'] == -1


def test_settings_yml_list():
    ctx = context_wrap(bi_conf_content)
    result = FakeYamlParser(ctx)
    assert result.data['remote_branch'] == -1
    assert result.data['remote_leaf'] == -1



from insights.parsers.cpuinfo import CpuInfo
from insights.tests import context_wrap

CPUINFO = """
COMMAND> cat /proc/cpuinfo
processor       : 0
vendor_id       : GenuineIntel
cpu family      : 6
model           : 45
model name      : Intel(R) Xeon(R) CPU E5-2690 0 @ 2.90GHz
stepping        : 2
microcode       : 1808
cpu MHz         : 2900.000
cache size      : 20480 KB
physical id     : 0
siblings        : 1
core id         : 0
cpu cores       : 1
flags           : fpu vme de pse tsc msr pae mce cx8 apic sep mtrr pge mca cmov pat pse36 clflush dts acpi mmx fxsr sse sse2 ss ht tm pbe syscall nx pdpe1gb rdtscp lm constant_tsc art arch_perfmon pebs bts rep_good nopl xtopology nonstop_tsc aperfmperf eagerfpu pni pclmulqdq dtes64 monitor ds_cpl vmx smx est tm2 ssse3 fma cx16 xtpr pdcm pcid sse4_1 sse4_2 x2apic movbe popcnt tsc_deadline_timer aes xsave avx f16c rdrand lahf_lm abm 3dnowprefetch epb intel_pt tpr_shadow vnmi flexpriority ept vpid fsgsbase tsc_adjust bmi1 hle avx2 smep bmi2 erms invpcid rtm mpx rdseed adx smap clflushopt xsaveopt xsavec xgetbv1 dtherm ida arat pln pts hwp hwp_notify hwp_act_window hwp_epp
address sizes   : 40 bits physical, 48 bits virtual

processor       : 1
vendor_id       : GenuineIntel
cpu family      : 6
model           : 45
model name      : Intel(R) Xeon(R) CPU E5-2690 0 @ 2.90GHz
stepping        : 2
microcode       : 1808
cpu MHz         : 2900.000
cache size      : 20480 KB
physical id     : 2
siblings        : 1
core id         : 0
cpu cores       : 1
flags           : fpu vme de pse tsc msr pae mce cx8 apic sep mtrr pge mca cmov pat pse36 clflush dts acpi mmx fxsr sse sse2 ss ht tm pbe syscall nx pdpe1gb rdtscp lm constant_tsc art arch_perfmon pebs bts rep_good nopl xtopology nonstop_tsc aperfmperf eagerfpu pni pclmulqdq dtes64 monitor ds_cpl vmx smx est tm2 ssse3 fma cx16 xtpr pdcm pcid sse4_1 sse4_2 x2apic movbe popcnt tsc_deadline_timer aes xsave avx f16c rdrand lahf_lm abm 3dnowprefetch epb intel_pt tpr_shadow vnmi flexpriority ept vpid fsgsbase tsc_adjust bmi1 hle avx2 smep bmi2 erms invpcid rtm mpx rdseed adx smap clflushopt xsaveopt xsavec xgetbv1 dtherm ida arat pln pts hwp hwp_notify hwp_act_window hwp_epp
address sizes   : 40 bits physical, 48 bits virtual

""".strip()

ONE_SOCKET_CPUINFO = """
processor	: 0
vendor_id	: GenuineIntel
cpu family	: 6
model		: 60
model name	: Intel(R) Core(TM) i7-4800MQ CPU @ 2.70GHz
stepping	: 3
microcode	: 0x1c
cpu MHz		: 3154.042
cache size	: 6144 KB
physical id	: 0
siblings	: 8
core id		: 0
cpu cores	: 4
apicid		: 0
initial apicid	: 0
fpu		: yes
fpu_exception	: yes
cpuid level	: 13
wp		: yes
flags		: fpu vme de pse tsc msr pae mce cx8 apic sep mtrr pge mca cmov pat pse36 clflush dts acpi mmx fxsr sse sse2 ss ht tm pbe syscall nx pdpe1gb rdtscp lm constant_tsc arch_perfmon pebs bts rep_good nopl xtopology nonstop_tsc aperfmperf eagerfpu pni pclmulqdq dtes64 monitor ds_cpl vmx smx est tm2 ssse3 fma cx16 xtpr pdcm pcid sse4_1 sse4_2 x2apic movbe popcnt tsc_deadline_timer aes xsave avx f16c rdrand lahf_lm abm ida arat epb pln pts dtherm tpr_shadow vnmi flexpriority ept vpid fsgsbase tsc_adjust bmi1 avx2 smep bmi2 erms invpcid xsaveopt
bogomips	: 5388.09
clflush size	: 64
cache_alignment	: 64
address sizes	: 39 bits physical, 48 bits virtual
power management:

processor	: 1
vendor_id	: GenuineIntel
cpu family	: 6
model		: 60
model name	: Intel(R) Core(TM) i7-4800MQ CPU @ 2.70GHz
stepping	: 3
microcode	: 0x1c
cpu MHz		: 3160.371
cache size	: 6144 KB
physical id	: 0
siblings	: 8
core id		: 0
cpu cores	: 4
apicid		: 1
initial apicid	: 1
fpu		: yes
fpu_exception	: yes
cpuid level	: 13
wp		: yes
flags		: fpu vme de pse tsc msr pae mce cx8 apic sep mtrr pge mca cmov pat pse36 clflush dts acpi mmx fxsr sse sse2 ss ht tm pbe syscall nx pdpe1gb rdtscp lm constant_tsc arch_perfmon pebs bts rep_good nopl xtopology nonstop_tsc aperfmperf eagerfpu pni pclmulqdq dtes64 monitor ds_cpl vmx smx est tm2 ssse3 fma cx16 xtpr pdcm pcid sse4_1 sse4_2 x2apic movbe popcnt tsc_deadline_timer aes xsave avx f16c rdrand lahf_lm abm ida arat epb pln pts dtherm tpr_shadow vnmi flexpriority ept vpid fsgsbase tsc_adjust bmi1 avx2 smep bmi2 erms invpcid xsaveopt
bogomips	: 5388.09
clflush size	: 64
cache_alignment	: 64
address sizes	: 39 bits physical, 48 bits virtual
power management:

processor	: 2
vendor_id	: GenuineIntel
cpu family	: 6
model		: 60
model name	: Intel(R) Core(TM) i7-4800MQ CPU @ 2.70GHz
stepping	: 3
microcode	: 0x1c
cpu MHz		: 3149.929
cache size	: 6144 KB
physical id	: 0
siblings	: 8
core id		: 1
cpu cores	: 4
apicid		: 2
initial apicid	: 2
fpu		: yes
fpu_exception	: yes
cpuid level	: 13
wp		: yes
flags		: fpu vme de pse tsc msr pae mce cx8 apic sep mtrr pge mca cmov pat pse36 clflush dts acpi mmx fxsr sse sse2 ss ht tm pbe syscall nx pdpe1gb rdtscp lm constant_tsc arch_perfmon pebs bts rep_good nopl xtopology nonstop_tsc aperfmperf eagerfpu pni pclmulqdq dtes64 monitor ds_cpl vmx smx est tm2 ssse3 fma cx16 xtpr pdcm pcid sse4_1 sse4_2 x2apic movbe popcnt tsc_deadline_timer aes xsave avx f16c rdrand lahf_lm abm ida arat epb pln pts dtherm tpr_shadow vnmi flexpriority ept vpid fsgsbase tsc_adjust bmi1 avx2 smep bmi2 erms invpcid xsaveopt
bogomips	: 5388.09
clflush size	: 64
cache_alignment	: 64
address sizes	: 39 bits physical, 48 bits virtual
power management:

processor	: 3
vendor_id	: GenuineIntel
cpu family	: 6
model		: 60
model name	: Intel(R) Core(TM) i7-4800MQ CPU @ 2.70GHz
stepping	: 3
microcode	: 0x1c
cpu MHz		: 3149.929
cache size	: 6144 KB
physical id	: 0
siblings	: 8
core id		: 1
cpu cores	: 4
apicid		: 3
initial apicid	: 3
fpu		: yes
fpu_exception	: yes
cpuid level	: 13
wp		: yes
flags		: fpu vme de pse tsc msr pae mce cx8 apic sep mtrr pge mca cmov pat pse36 clflush dts acpi mmx fxsr sse sse2 ss ht tm pbe syscall nx pdpe1gb rdtscp lm constant_tsc arch_perfmon pebs bts rep_good nopl xtopology nonstop_tsc aperfmperf eagerfpu pni pclmulqdq dtes64 monitor ds_cpl vmx smx est tm2 ssse3 fma cx16 xtpr pdcm pcid sse4_1 sse4_2 x2apic movbe popcnt tsc_deadline_timer aes xsave avx f16c rdrand lahf_lm abm ida arat epb pln pts dtherm tpr_shadow vnmi flexpriority ept vpid fsgsbase tsc_adjust bmi1 avx2 smep bmi2 erms invpcid xsaveopt
bogomips	: 5388.09
clflush size	: 64
cache_alignment	: 64
address sizes	: 39 bits physical, 48 bits virtual
power management:

processor	: 4
vendor_id	: GenuineIntel
cpu family	: 6
model		: 60
model name	: Intel(R) Core(TM) i7-4800MQ CPU @ 2.70GHz
stepping	: 3
microcode	: 0x1c
cpu MHz		: 3153.937
cache size	: 6144 KB
physical id	: 0
siblings	: 8
core id		: 2
cpu cores	: 4
apicid		: 4
initial apicid	: 4
fpu		: yes
fpu_exception	: yes
cpuid level	: 13
wp		: yes
flags		: fpu vme de pse tsc msr pae mce cx8 apic sep mtrr pge mca cmov pat pse36 clflush dts acpi mmx fxsr sse sse2 ss ht tm pbe syscall nx pdpe1gb rdtscp lm constant_tsc arch_perfmon pebs bts rep_good nopl xtopology nonstop_tsc aperfmperf eagerfpu pni pclmulqdq dtes64 monitor ds_cpl vmx smx est tm2 ssse3 fma cx16 xtpr pdcm pcid sse4_1 sse4_2 x2apic movbe popcnt tsc_deadline_timer aes xsave avx f16c rdrand lahf_lm abm ida arat epb pln pts dtherm tpr_shadow vnmi flexpriority ept vpid fsgsbase tsc_adjust bmi1 avx2 smep bmi2 erms invpcid xsaveopt
bogomips	: 5388.09
clflush size	: 64
cache_alignment	: 64
address sizes	: 39 bits physical, 48 bits virtual
power management:

processor	: 5
vendor_id	: GenuineIntel
cpu family	: 6
model		: 60
model name	: Intel(R) Core(TM) i7-4800MQ CPU @ 2.70GHz
stepping	: 3
microcode	: 0x1c
cpu MHz		: 3160.371
cache size	: 6144 KB
physical id	: 0
siblings	: 8
core id		: 2
cpu cores	: 4
apicid		: 5
initial apicid	: 5
fpu		: yes
fpu_exception	: yes
cpuid level	: 13
wp		: yes
flags		: fpu vme de pse tsc msr pae mce cx8 apic sep mtrr pge mca cmov pat pse36 clflush dts acpi mmx fxsr sse sse2 ss ht tm pbe syscall nx pdpe1gb rdtscp lm constant_tsc arch_perfmon pebs bts rep_good nopl xtopology nonstop_tsc aperfmperf eagerfpu pni pclmulqdq dtes64 monitor ds_cpl vmx smx est tm2 ssse3 fma cx16 xtpr pdcm pcid sse4_1 sse4_2 x2apic movbe popcnt tsc_deadline_timer aes xsave avx f16c rdrand lahf_lm abm ida arat epb pln pts dtherm tpr_shadow vnmi flexpriority ept vpid fsgsbase tsc_adjust bmi1 avx2 smep bmi2 erms invpcid xsaveopt
bogomips	: 5388.09
clflush size	: 64
cache_alignment	: 64
address sizes	: 39 bits physical, 48 bits virtual
power management:

processor	: 6
vendor_id	: GenuineIntel
cpu family	: 6
model		: 60
model name	: Intel(R) Core(TM) i7-4800MQ CPU @ 2.70GHz
stepping	: 3
microcode	: 0x1c
cpu MHz		: 3154.042
cache size	: 6144 KB
physical id	: 0
siblings	: 8
core id		: 3
cpu cores	: 4
apicid		: 6
initial apicid	: 6
fpu		: yes
fpu_exception	: yes
cpuid level	: 13
wp		: yes
flags		: fpu vme de pse tsc msr pae mce cx8 apic sep mtrr pge mca cmov pat pse36 clflush dts acpi mmx fxsr sse sse2 ss ht tm pbe syscall nx pdpe1gb rdtscp lm constant_tsc arch_perfmon pebs bts rep_good nopl xtopology nonstop_tsc aperfmperf eagerfpu pni pclmulqdq dtes64 monitor ds_cpl vmx smx est tm2 ssse3 fma cx16 xtpr pdcm pcid sse4_1 sse4_2 x2apic movbe popcnt tsc_deadline_timer aes xsave avx f16c rdrand lahf_lm abm ida arat epb pln pts dtherm tpr_shadow vnmi flexpriority ept vpid fsgsbase tsc_adjust bmi1 avx2 smep bmi2 erms invpcid xsaveopt
bogomips	: 5388.09
clflush size	: 64
cache_alignment	: 64
address sizes	: 39 bits physical, 48 bits virtual
power management:

processor	: 7
vendor_id	: GenuineIntel
cpu family	: 6
model		: 60
model name	: Intel(R) Core(TM) i7-4800MQ CPU @ 2.70GHz
stepping	: 3
microcode	: 0x1c
cpu MHz		: 3133.687
cache size	: 6144 KB
physical id	: 0
siblings	: 8
core id		: 3
cpu cores	: 4
apicid		: 7
initial apicid	: 7
fpu		: yes
fpu_exception	: yes
cpuid level	: 13
wp		: yes
flags		: fpu vme de pse tsc msr pae mce cx8 apic sep mtrr pge mca cmov pat pse36 clflush dts acpi mmx fxsr sse sse2 ss ht tm pbe syscall nx pdpe1gb rdtscp lm constant_tsc arch_perfmon pebs bts rep_good nopl xtopology nonstop_tsc aperfmperf eagerfpu pni pclmulqdq dtes64 monitor ds_cpl vmx smx est tm2 ssse3 fma cx16 xtpr pdcm pcid sse4_1 sse4_2 x2apic movbe popcnt tsc_deadline_timer aes xsave avx f16c rdrand lahf_lm abm ida arat epb pln pts dtherm tpr_shadow vnmi flexpriority ept vpid fsgsbase tsc_adjust bmi1 avx2 smep bmi2 erms invpcid xsaveopt
bogomips	: 5388.09
clflush size	: 64
cache_alignment	: 64
address sizes	: 39 bits physical, 48 bits virtual
power management:

"""


def test_cpuinfo():
    cpu_info = CpuInfo(context_wrap(CPUINFO))
    assert cpu_info.cpu_count == 2
    assert cpu_info.socket_count == 2
    assert cpu_info.vendor == "GenuineIntel"
    assert cpu_info.model_name == "Intel(R) Xeon(R) CPU E5-2690 0 @ 2.90GHz"
    assert cpu_info.get_processor_by_index(0) == {
        "cpus": "0",
        "sockets": "0",
        "vendors": "GenuineIntel",
        "models": "Intel(R) Xeon(R) CPU E5-2690 0 @ 2.90GHz",
        "model_ids": "45",
        "families": "6",
        "clockspeeds": "2900.000",
        "cache_sizes": "20480 KB",
        "cpu_cores": "1",
        "flags": "fpu vme de pse tsc msr pae mce cx8 apic sep mtrr pge mca cmov pat pse36 clflush dts acpi mmx fxsr sse sse2 ss ht tm pbe syscall nx pdpe1gb rdtscp lm constant_tsc art arch_perfmon pebs bts rep_good nopl xtopology nonstop_tsc aperfmperf eagerfpu pni pclmulqdq dtes64 monitor ds_cpl vmx smx est tm2 ssse3 fma cx16 xtpr pdcm pcid sse4_1 sse4_2 x2apic movbe popcnt tsc_deadline_timer aes xsave avx f16c rdrand lahf_lm abm 3dnowprefetch epb intel_pt tpr_shadow vnmi flexpriority ept vpid fsgsbase tsc_adjust bmi1 hle avx2 smep bmi2 erms invpcid rtm mpx rdseed adx smap clflushopt xsaveopt xsavec xgetbv1 dtherm ida arat pln pts hwp hwp_notify hwp_act_window hwp_epp"
    }
    assert cpu_info.cpu_speed == "2900.000"
    assert cpu_info.cache_size == "20480 KB"
    assert cpu_info.model_number == "45"
    assert "mmx" in cpu_info.flags
    assert "avx512f" not in cpu_info.flags
    for i, cpu in enumerate(cpu_info):
        assert cpu['cpus'] == str(i)


def test_one_socket_cpuinfo():
    cpu_info = CpuInfo(context_wrap(ONE_SOCKET_CPUINFO))
    assert cpu_info.cpu_count == 8
    assert cpu_info.socket_count == 1
    assert cpu_info.core_total == 32


def test_empty_cpuinfo():
    cpu_info = CpuInfo(context_wrap(""))
    assert cpu_info.cpu_count == 0



"""
Kernel system control information
=================================

Shared parsers for parsing file ``/etc/sysctl.conf`` and command ``sysctl -a``.

Parsers included in this module are:

Sysctl - command ``sysctl -a``
------------------------------

SysctlConf - file ``/etc/sysctl.conf``
--------------------------------------

SysctlConfInitramfs - command ``lsinitrd``
------------------------------------------

"""
from .. import Parser, LogFileOutput, parser, LegacyItemAccess, CommandParser
from ..parsers import split_kv_pairs
from insights.specs import Specs


@parser(Specs.sysctl_conf)
class SysctlConf(Parser):
    """Parse `/etc/sysctl.conf` file

    Sample input::

        # sysctl.conf sample
        #
          kernel.domainname = example.com

        ; this one has a space which will be written to the sysctl!
          kernel.modprobe = /sbin/mod probe

    Attributes
    ----------
    data: OrderedDict
        Dictionary containing key/value pairs for the lines in the
        configuration file.  Dictionary is in order keywords first
        appear in the lines.

    Examples
    --------
    >>> shared[SysctlConf].data['kernel.domainname']
    'example.com'
    >>> shared[SysctlConf].data['kernel.modprobe']
    '/sbin/mod probe'
    """

    def parse_content(self, content):
        # Valid comments are both # and ; so remove one locally,
        # other comments and blank lines are removed by split fxn.
        lines = [l for l in content if not l.strip().startswith(';')]
        self.data = split_kv_pairs(lines, ordered=True)


@parser(Specs.sysctl)
class Sysctl(LegacyItemAccess, CommandParser):
    """Parse the output of `sysctl -a` command.

    Sample input::

        kernel.domainname = example.com
        kernel.modprobe = /sbin/modprobe

    Examples
    --------
    >>> shared[Sysctl]['kernel.domainname']
    'example.com'
    >>> shared[Sysctl].get('kernel.modprobe')
    '/sbin/modprobe'
    >>> 'kernel.modules_disabled' in shared[Sysctl]
    False
    """

    def parse_content(self, content):
        self.data = split_kv_pairs(content)


@parser(Specs.sysctl_conf_initramfs)
class SysctlConfInitramfs(CommandParser, LogFileOutput):
    """Shared parser for the output of ``lsinitrd`` applied to kdump
    initramfs images to view ``sysctl.conf`` and ``sysctl.d``
    configurations.

    For now, the file is treated as raw lines (as a ``LogFileOutput``
    parser.  This is because the output of the command, applied to
    multiple files to examine multiple files does not seem to be
    unambiguously parsible.

    Since the only plugins requiring the file to date "grep out"
    certain strings, this approach will suffice.

    .. note::
        Please refer to its super-class :class:`insights.core.LogFileOutput`
    """
    pass



"""
Kernel dump configuration files
===============================

This module contains the following parsers:

KDumpConf - file ``/etc/kdump.conf``
------------------------------------

KexecCrashLoaded - file ``/sys/kernel/kexec_crash_loaded``
----------------------------------------------------------

KexecCrashSize - file ``/sys/kernel/kexec_crash_size``
------------------------------------------------------

"""

import re

try:
    from urlparse import urlparse
except ImportError:
    from urllib.parse import urlparse

from insights import Parser, parser, LegacyItemAccess
from insights.parsers import ParseException
from insights.specs import Specs


@parser(Specs.kdump_conf)
class KDumpConf(Parser, LegacyItemAccess):
    """
    A dictionary like object for the values of the ``/etc/kdump.conf`` file.

    Attributes:

        lines (list): raw lines from the file, in order
        data (dict): a dictionary of options set in the data
        comments(list): fully commented lines
        inline_comments(list): lines containing inline comments
        target(tuple): target line parsed as a (x, y) tuple if set, else None

    The ``data`` property has two special behaviours:

    * If an option - e.g. ``blacklist`` - is repeated, its values are
      collected together in a list.  Options that only appear once have
      their values stored as is.
    * The ``options`` option is special - it appears in the form ``option
      module value``.  The ``options`` key in the data dictionary is therefore
      stored as a dictionary, keyed on the ``module`` name.

    The ``target`` property has following possibilities:

    * If target-line starts with any keyword in ['raw', 'ssh', 'net', 'nfs', 'nfs4'],
      return tuple (keyword, value).
    * If target-line is set with '<fs_type> <partation>',
      return tuple (<fs_type>, <partation>).
    * If target-line is not set, the target is default which is depending on
      what's mounted in the current system, return None instead of tuple here.

    Main helper functions:

    * ``options`` - the ``options`` value in the data(see above).

    Sample ``/etc/kdump.conf`` file::

        path /var/crash
        core_collector makedumpfile -c --message-level 1 -d 24
        default shell

    Examples:

        >>> kd.is_local_disk
        True
        >>> kd.is_ssh()
        False
        >>> 'path' in kd
        True
    """
    NET_COMMANDS = set(['nfs', 'net', 'ssh'])
    SUPPORTED_FS_TYPES = ['ext2', 'ext3', 'ext4', 'btrfs', 'xfs']

    def parse_content(self, content):
        lines = list(content)
        opt_kw = 'options'
        items = {opt_kw: {}}
        # Paul Wayper - 2017-03-27 - why do we care about comments?
        comments = []
        inline_comments = []

        for _line in content:
            line = _line.strip()
            if not line:
                continue
            # Ignore lines that are entirely comments
            if line.startswith('#'):
                comments.append(_line)
                continue
            # Remove comments
            if '#' in line:
                comment_start = line.index('#')
                inline_comments.append(_line)
                line = line[0:comment_start]

            # Settings of the form 'option value' where value is the rest of
            # the line.  No equals is expected here.
            lineparts = [s.strip() for s in line.split(None, 1)]
            # All options must have a value
            if len(lineparts) < 2:
                continue
            opt, value = (lineparts)

            if opt != opt_kw:
                # Some items can be repeated - if they are, create a list of
                # their values
                if opt in items:
                    # Append to the list, creating if necessary
                    if not isinstance(items[opt], list):
                        items[opt] = [items[opt]]
                    items[opt].append(value)
                else:
                    items[opt] = value
            else:
                # 'options' is special - it becomes a dictionary
                mod, rest = value.split(None, 1)
                items[opt_kw][mod] = rest.strip()

        self.lines = lines
        self.data = items
        self.comments = comments
        self.inline_comments = inline_comments
        self.target = self._parse_target()

    def options(self, module):
        """
        Returns the options for this module in the settings.

        Arguments:
            module(str): The module name

        Returns:
            (str) The module's options, or '' if either ``options`` or
              ``module`` is not found.
        """
        return self.get('options', {}).get(module, '')

    def _network_lines(self, net_commands=NET_COMMANDS):
        """
        A list of all the options in the given list of commands, defaulting
        to the list of network destinations for kernel dumps (i.e. 'ssh',
        'nfs', 'nfs4' and 'net').
        """
        return filter(None, [self.get(n) for n in net_commands])

    def get_ip(self, net_commands=NET_COMMANDS):
        """
        Find the first IP address in the given list of commands.  Uses
        ``_network_lines`` above to find the list of commands.  The first
        line that lists an IP address is returned, otherwise None is returned.
        """
        ip_re = re.compile(r'(\d{1,3}\.){3}\d{1,3}')
        for l in self._network_lines(net_commands):
            matched_ip = ip_re.search(l)
            if matched_ip:
                return matched_ip.group()

    def is_ssh(self):
        """
        Is the destination of the kernel dump an ssh connection?
        """
        return 'ssh' in self or ('net' in self and '@' in self['net'])

    def is_nfs(self):
        """
        Is the destination of the kernel dump a NFS or NFSv4 connection?
        """
        return (
            ('nfs' in self or 'nfs4' in self) or
            ('net' in self and '@' not in self['net'])
        )

    def get_hostname(self, net_commands=NET_COMMANDS):
        """
        Find the first host name in the given list of commands.  Uses
        ``_network_lines`` above to find the list of commands.  The first
        line that matches ``urlparse``'s definition of a host name is
        returned, or None is returned.
        """
        for l in self._network_lines(net_commands):
            # required for urlparse to interpret as host instead of
            # relative path
            if '//' not in l:
                l = '//' + l
            netloc = urlparse(l).netloc

            # strip user:pass@
            i = netloc.find('@')
            if i != -1:
                netloc = netloc[i + 1:]

            # strip port
            return netloc.rsplit(':', 1)[0]

    @property
    def ip(self):
        """
        Uses get_ip() above to give the first IP address found in the list of
        crash dump destinations.
        """
        return self.get_ip()

    @property
    def hostname(self):
        """
        Uses get_hostname() above to give the first host name found in the
        list of crash dump destinations.
        """
        return self.get_hostname()

    @property
    def using_local_disk(self):
        """
        Is kdump configured to only use local disk?

        Several target types:

        * If 'raw' is given, then the dump is local.
        * If 'ssh', 'net', 'nfs', or 'nfs4' is given, then the dump is NOT local.
        * If '<fs type> <partition>' is given, then the dump is local.
        * Otherwise, the dump is local.

        Since only one target could be set, the logic used here is checking
        if remote target is used, return True for not.
        """
        return not ('ssh' in self.data or 'net' in self.data or
                    'nfs' in self.data or 'nfs4' in self.data)

    def _parse_target(self):
        """
        More than one dump targets will lead to kudmp service start failure.
        Raise an exception here if more than one target is set here.
        https://access.redhat.com/documentation/en-us/red_hat_enterprise_linux/7/html/kernel_administration_guide/kernel_crash_dump_guide#sect-supported-kdump-targets
        """
        target = None
        keys = ['ssh', 'net', 'nfs', 'nfs4', 'raw'] + self.SUPPORTED_FS_TYPES
        for k in keys:
            if k in self.data:
                v = self.data[k]
                if isinstance(v, list):
                    raise ParseException("More than one %s type targets are\
                                         configured." % k)
                if target:
                    raise ParseException("More than one target is configured.")
                else:
                    target = (k, v)
        return target


@parser(Specs.kexec_crash_loaded)
class KexecCrashLoaded(Parser):
    """
    A simple parser to determine if a crash kernel (i.e. a second kernel
    capable of capturing the machine state should the main kernel crash) is
    present.

    This simply returns a set of whether the ``/sys/kernel/kexec_crash_loaded``
    file has the value ``1``.
    """

    def parse_content(self, content):
        if len(content) == 0:
            self.is_loaded = False
            return
        line = list(content)[0].strip()
        self.is_loaded = line == '1'


@parser(Specs.kexec_crash_size)
class KexecCrashSize(Parser):
    """
    Parses the `/sys/kernel/kexec_crash_size` file which tells the
    reserved memory size for the crash kernel.

    Attributes:
        size (int): reserved memory size for the crash kernel, or 0 if not found.
    """

    def parse_content(self, content):
        self.size = 0
        if len(content) == 0:
            return
        size = list(content)[0].strip()
        if size.isdigit():
            self.size = int(size)



"""
Nmcli parsers
=============

This file will parse the output of command line tools used to manage
NetworkManager.

Parsers provided by this module include:

NmcliDevShow - command ``/usr/bin/nmcli dev show``
--------------------------------------------------
NmcliConnShow - command ''/usr/bin/nmcli conn show``
----------------------------------------------------
"""


import re
from .. import parser, LegacyItemAccess, get_active_lines, CommandParser
from insights.specs import Specs
from insights.parsers import parse_delimited_table


@parser(Specs.nmcli_dev_show)
class NmcliDevShow(CommandParser, LegacyItemAccess):
    """
    This class will parse the output of command ``nmcli dev show``, and the information
    will be stored in dictionary format.

    NetworkManager displays all the devices and their current states along with network
    configuration and connection status.

    Attributes:
        data (dict): Dictionary of keys with values in dict.
        connected_devices(list): list of devices who's state is connected.

    Sample input for ``/usr/bin/nmcli dev show``::

        GENERAL.DEVICE:                         em3
        GENERAL.TYPE:                           ethernet
        GENERAL.HWADDR:                         B8:2A:72:DE:F8:B9
        GENERAL.MTU:                            1500
        GENERAL.STATE:                          100 (connected)
        GENERAL.CONNECTION:                     em3
        GENERAL.CON-PATH:                       /org/freedesktop/NetworkManager/ActiveConnection/1
        WIRED-PROPERTIES.CARRIER:               on
        IP4.ADDRESS[1]:                         10.16.184.98/22
        IP4.GATEWAY:                            10.16.187.254
        IP4.DNS[1]:                             10.16.36.29
        IP4.DNS[2]:                             10.11.5.19
        IP4.DNS[3]:                             10.5.30.160
        IP4.DOMAIN[1]:                          khw.lab.eng.bos.example.com
        IP6.ADDRESS[1]:                         2620:52:0:10bb:ba2a:72ff:fede:f8b9/64
        IP6.ADDRESS[2]:                         fe80::ba2a:72ff:fede:f8b9/64
        IP6.GATEWAY:                            fe80:52:0:10bb::fc
        IP6.ROUTE[1]:                           dst = 2620:52:0:10bb::/64, nh = ::, mt = 100

        GENERAL.DEVICE:                         em1
        GENERAL.TYPE:                           ethernet
        GENERAL.HWADDR:                         B8:2A:72:DE:F8:BB
        GENERAL.MTU:                            1500
        GENERAL.STATE:                          30 (disconnected)
        GENERAL.CONNECTION:                     --
        GENERAL.CON-PATH:                       --
        WIRED-PROPERTIES.CARRIER:               off

        GENERAL.DEVICE:                         em2
        GENERAL.TYPE:                           ethernet
        GENERAL.HWADDR:                         B8:2A:72:DE:F8:BC
        GENERAL.MTU:                            1500
        GENERAL.STATE:                          30 (disconnected)
        GENERAL.CONNECTION:                     --
        GENERAL.CON-PATH:                       --
        WIRED-PROPERTIES.CARRIER:               off

    Sample Output::

        {
            'em3':
                {
                  'IP4_DNS3': '10.5.30.160',
                  'IP4_DNS2': '10.11.5.19',
                  'IP4_DNS1': '10.16.36.29',
                  'IP6_ADDRESS2': 'fe80::ba2a:72ff:fede:f8b9/64',
                  'CONNECTION': 'em3',
                  'IP6_ADDRESS1': '2620:52:0:10bb:ba2a:72ff:fede:f8b9/64',
                  'CON-PATH': '/org/freedesktop/NetworkManager/ActiveConnection/1',
                  'IP4_ADDRESS1': '10.16.184.98/22',
                  'MTU': '1500',
                  'IP4_GATEWAY': '10.16.187.254',
                  'STATE': 'connected',
                  'CARRIER': 'on',
                  'IP4_DOMAIN1': 'khw.lab.eng.bos.example.com',
                  'IP6_ROUTE1': 'dst = 2620:52:0:10bb::/64, nh = ::, mt = 100',
                  'HWADDR': 'B8:2A:72:DE:F8:B9',
                  'IP6_GATEWAY': 'fe80:52:0:10bb::fc',
                  'TYPE': 'ethernet'
                },
              'em2':
                {
                  'STATE': 'connected',
                  'CARRIER': 'off',
                  'HWADDR': 'B8:2A:72:DE:F8:BC',
                  'CON-PATH': '--',
                  ...
                  ...
                }
            ...
            ...
        }

    Examples:
        >>> type(nmcli_obj)
        <class 'insights.parsers.nmcli.NmcliDevShow'>
        >>> nmcli_obj.data['em3']['STATE']
        'connected'
        >>> nmcli_obj.data['em2']['HWADDR']
        'B8:2A:72:DE:F8:BC'
        >>> sorted(nmcli_obj.connected_devices)
        ['em1', 'em2', 'em3']

    """

    def parse_content(self, content):
        self.data = {}
        per_device = {}
        current_dev = ""
        for line in get_active_lines(content):
            if not ("not found" in line or "Error" in line or "No such file" in line):
                key, val = line.split(": ")
                if "IP" in key:
                    proto = re.sub(r'\[|\]', r'', key.split('.')[1])
                    key = key.split('.')[0] + "_" + proto
                else:
                    key = key.split('.')[1]
                val = re.sub(r'\d+\s|\(|\)', r'', val.strip())

                # Device configuration details starts here
                if key == "DEVICE" and not current_dev:
                    current_dev = val
                    continue
                elif key == "DEVICE" and current_dev:
                    self.data[current_dev] = per_device
                    current_dev = val
                    per_device = {}
                    continue
                per_device.update({key: val})
        if current_dev and per_device:
            # Last device configuration details
            self.data[current_dev] = per_device

    @property
    def connected_devices(self):
        """(list): The list of devices who's state is connected and managed by NetworkManager"""
        con_dev = []
        for key in self.data:
            if 'STATE' in self.data[key] and self.data[key]['STATE'] == 'connected':
                con_dev.append(key)
        return con_dev


@parser(Specs.nmcli_conn_show)
class NmcliConnShow(CommandParser):
    """
    This file will parse the output of all the nmcli connections.

    Sample configuration from a teamed interface in file ``/usr/bin/nmcli conn show``::

       NAME      UUID                                  TYPE      DEVICE
       enp0s3    320d4923-c410-4b22-b7e9-afc5f794eecc  ethernet  enp0s3
       virbr0    7c7dec66-4a8c-4b49-834a-889194b3b83c  bridge    virbr0
       test-net  f858b1cc-d149-4de0-93bc-b1826256847a  ethernet  --

    Examples:
        >>> type(static_conn)
        <class 'insights.parsers.nmcli.NmcliConnShow'>
        >>> static_conn.disconnected_connection
        ['test-net-1']

    Attributes:
        data (list): list of connections wrapped in dictionaries

    """
    def parse_content(self, content):
        self.data = parse_delimited_table(content)
        self._disconnected_connection = []
        for all_connection in self.data:
            if all_connection['DEVICE'] == "--":
                self._disconnected_connection.append(all_connection['NAME'])

    @property
    def disconnected_connection(self):
        """(list): It will return all the disconnected static route connections."""
        return self._disconnected_connection



from insights.core.context import OSP
from insights.parsers.rabbitmq import RabbitMQReport
from insights.tests import context_wrap

osp_controller = OSP()
osp_controller.role = "Controller"


RABBITMQCTL_REPORT_0 = """
Error: unable to connect to node rabbit@dprcclab002: nodedown

DIAGNOSTICS
===========

attempted to contact: [rabbit@dprcclab002]

rabbit@dprcclab002:
  * connected to epmd (port 4369) on dprcclab002
  * epmd reports: node 'rabbit' not running at all
                  no other nodes on dprcclab002
  * suggestion: start the node

current node details:
- node name: 'rabbitmq-cli-08@dprcclab002'
- home dir: /var/lib/rabbitmq
- cookie hash: k0XaAkNZ29GQG22RA/Uyxw==
"""

RABBITMQCTL_REPORT_1 = """
Reporting server status on {{2016,10,17},{3,42,41}}

 ...
Status of node rabbit@rabbitmq0 ...
[{pid,5085},
 {os,{unix,linux}},
 {erlang_version,
     "Erlang R16B03-1 (erts-5.10.4) [source] [64-bit] [smp:4:4] [async-threads:30] [hipe] [kernel-poll:true]\n"},
 {listeners,[{clustering,25672,"::"},{amqp,5672,"::"}]},
 {vm_memory_high_watermark,0.4},
 {vm_memory_limit,1590132736},
 {disk_free_limit,50000000},
 {disk_free,40712921088},
 {file_descriptors,
     [{total_limit,924},{total_used,3},{sockets_limit,829},{sockets_used,1}]},
 {processes,[{limit,1048576},{used,180}]},
 {run_queue,0},
 {uptime,5590}]

Cluster status of node rabbit@rabbitmq0 ...
[{nodes,[{disc,[rabbit@rabbitmq0]}]},

Application environment of node rabbit@rabbitmq0 ...
[{auth_backends,[rabbit_auth_backend_internal]},

Permissions on /:
user    configure       write   read
guest   .*      .*      .*
redhat  redhat.*        .*      .*
redhat1 redhat.*        .*      .*

Permissions on test_vhost:

Policies on /:

Policies on test_vhost:

...done.
"""

RABBITMQCTL_REPORT_2 = """
Reporting server status on {{2016,4,26},{6,51,27}}

 ...
Status of node 'rabbit@overcloud-controller-2' ...
[{pid,4087},
 {vm_memory_limit,26943271731},
 {disk_free_limit,50000000},
 {disk_free,257304219648},
 {file_descriptors,[{total_limit,3996},
                    {total_used,835},
                    {sockets_limit,3594},
                    {sockets_used,833}]},
 {uptime,3075474}]

Cluster status of node 'rabbit@overcloud-controller-2' ...
[{nodes,[{disc,['rabbit@overcloud-controller-0',
                'rabbit@overcloud-controller-1',
                'rabbit@overcloud-controller-2']}]},
 {partitions,[]}]

Application environment of node 'rabbit@overcloud-controller-2' ...
[{auth_backends,[rabbit_auth_backend_internal]}]

Status of node 'rabbit@overcloud-controller-1' ...
[{pid,9304},
 {disk_free_limit,50000000},
 {disk_free,260561866752},
 {file_descriptors,[{total_limit,3996},
                    {total_used,853},
                    {sockets_limit,3594},
                    {sockets_used,851}]},
 {uptime,3075482}]

Cluster status of node 'rabbit@overcloud-controller-2' ...
[{nodes,[{disc,['rabbit@overcloud-controller-0',
                'rabbit@overcloud-controller-1',
                'rabbit@overcloud-controller-2']}]},
 {partitions,[]}]

Application environment of node 'rabbit@overcloud-controller-2' ...
[{auth_backends,[rabbit_auth_backend_internal]}]

Status of node 'rabbit@overcloud-controller-0' ...
[{pid,6005},
 {disk_free_limit,50000000},
 {disk_free,259739344896},
 {file_descriptors,[{total_limit,3996},
                    {total_used,967},
                    {sockets_limit,3594},
                    {sockets_used,965}]},
 {uptime,3075485}]

Cluster status of node 'rabbit@overcloud-controller-2' ...
[{nodes,[{disc,['rabbit@overcloud-controller-0',
                'rabbit@overcloud-controller-1',
                'rabbit@overcloud-controller-2']}]},
 {partitions,[]}]

Application environment of node 'rabbit@overcloud-controller-2' ...
[{auth_backends,[rabbit_auth_backend_internal]}]

Permissions on /:
user	configure	write	read
guest	.*	.*	.*

Policies on /:
vhost	name	apply-to	pattern	definition	priority
/	ha-all	all	^(?!amq\\.).*	{"ha-mode":"all"}	0

Parameters on /:

...done.
"""


def test_rabbitmq_report():
    result = RabbitMQReport(context_wrap(RABBITMQCTL_REPORT_0,
            hostname="controller_1", osp=osp_controller)).result
    assert result is None

    result = RabbitMQReport(context_wrap(RABBITMQCTL_REPORT_1,
            hostname="controller_1", osp=osp_controller)).result
    assert result.get("nstat").get("rabbit@rabbitmq0").\
            get("file_descriptors").get("total_limit") == "924"
    permissions = {'/': {'redhat1': ['redhat.*', '.*', '.*'],
                        'guest': ['.*', '.*', '.*'],
                        'redhat': ['redhat.*', '.*', '.*']},
                   'test_vhost': ''}
    assert result.get("perm") == permissions

    result = RabbitMQReport(context_wrap(RABBITMQCTL_REPORT_2,
            hostname="controller_1", osp=osp_controller)).result
    assert result.get("nstat").get("'rabbit@overcloud-controller-2'").\
            get("file_descriptors").get("total_limit") == '3996'
    assert len(result.get("nstat")) == 3
    permissions = {'/': {'guest': ['.*', '.*', '.*']}}
    assert result.get("perm") == permissions




"""
IronicConf - file ``/etc/ironic/ironic.conf``
=============================================
This class provides parsing for the file ``/etc/ironic/ironic.conf``.
See the ``IniConfigFile`` class for more usage information.
"""


from .. import IniConfigFile, parser, add_filter
from insights.specs import Specs

add_filter(Specs.ironic_conf, ["["])


@parser(Specs.ironic_conf)
class IronicConf(IniConfigFile):
    """
    Ironic configuration parser class, based on the ``IniConfigFile`` class.

    Sample input data is in the format::

        [DEFAULT]

        auth_strategy=keystone
        default_resource_class=baremetal
        enabled_hardware_types=idrac,ilo,ipmi,redfish
        enabled_bios_interfaces=no-bios
        enabled_boot_interfaces=ilo-pxe,pxe
        enabled_console_interfaces=ipmitool-socat,ilo,no-console
        force_raw_images = True

        [agent]

        deploy_logs_collect=always
        deploy_logs_storage_backend=local
        deploy_logs_local_path=/var/log/ironic/deploy/

        [cinder]

        auth_url=http://1.1.1.1:5000
        project_domain_name=Default
        project_name=service
        user_domain_name=Default


    Examples:

        >>> ironic_conf.has_option("agent", "deploy_logs_collect")
        True
        >>> ironic_conf.get("DEFAULT", "auth_strategy") == "keystone"
        True

    """
    pass



"""
LibVirtdLog - file ``/var/log/libvirt/libvirtd.log``
====================================================

This is a fairly simple parser to read the logs of libvirtd.  It uses the
LogFileOutput parser class.

Sample input::

    2013-10-23 17:32:19.909+0000: 14069: debug : do_open:1174 : trying driver 1 (ESX) ...
    2013-10-23 17:32:19.909+0000: 14069: debug : do_open:1180 : driver 1 ESX returned DECLINED
    2013-10-23 17:32:19.909+0000: 14069: debug : do_open:1174 : trying driver 2 (remote) ...
    2013-10-23 17:32:19.957+0000: 14069: error : virNetTLSContextCheckCertDN:418 : Certificate [session] owner does not match the hostname AA.BB.CC.DD <============= IP Address
    2013-10-23 17:32:19.957+0000: 14069: warning : virNetTLSContextCheckCertificate:1102 : Certificate check failed Certificate [session] owner does not match the hostname AA.BB.CC.DD
    2013-10-23 17:32:19.957+0000: 14069: error : virNetTLSContextCheckCertificate:1105 : authentication failed: Failed to verify peer's certificate

Examples:

    >>> LibVirtdLog.filters.append('NetTLSContext')
    >>> LibVirtdLog.token_scan('check_failed', 'Certificate check failed')
    >>> virtlog = shared[LibVirtdLog]
    >>> len(virtlog.lines) # All lines, before filtering
    6
    >>> len(virtlog.get('NetTLSContext')) # After filtering
    3
    >>> virtlog.check_failed
    True

"""

from .. import LogFileOutput, parser
from insights.specs import Specs


@parser(Specs.libvirtd_log)
class LibVirtdLog(LogFileOutput):
    """
    Parse the ``/var/log/libvirt/libvirtd.log`` log file.

    .. note::
        Please refer to its super-class :class:`insights.core.LogFileOutput`
    """
    pass



from insights.configtree import PushBack, parse_string, typed, eat_whitespace
from insights.configtree import Directive, Root, Section


class LineCounter(PushBack):
    def __init__(self, *args, **kwargs):
        super(LineCounter, self).__init__(*args, **kwargs)
        self.lines = 0

    def push_back(self, item):
        if item == "\n":
            self.lines -= 1
        super(LineCounter, self).push_back(item)

    def push_all(self, items):
        for i in reversed(items):
            self.push_back(i)

    def next(self):
        v = super(LineCounter, self).next()
        if v == "\n":
            self.lines += 1
        return v

    __next__ = next


class DocParser(object):
    def __init__(self, ctx=None, line_end=";"):
        self.ctx = ctx
        self.line_end = line_end

    def parse_bare(self, pb):
        buf = []
        try:
            end = "{" + self.line_end
            while not pb.peek().isspace() and pb.peek() not in end:
                buf.append(next(pb))
        except StopIteration:
            pass
        return "".join(buf)

    def parse_attrs(self, pb):
        attrs = []
        eat_whitespace(pb, to=self.line_end)
        while pb.peek() not in ("{}" + self.line_end):
            if pb.peek() in ('"', "'"):
                attrs.append(parse_string(pb))
            else:
                attrs.append(self.parse_bare(pb))

            if pb.peek() != self.line_end:
                eat_whitespace(pb, to=self.line_end)

        if len(attrs) == 1:
            attrs = [typed(attrs[0])]

        if pb.peek() == self.line_end:
            next(pb)  # eat it
        eat_whitespace(pb)
        return attrs

    def parse_section_body(self, pb):
        next(pb)  # eat bracket
        eat_whitespace(pb)

        body = []
        while pb.peek() != "}":
            body.append(self.parse_statement(pb))
        next(pb)  # eat bracket
        eat_whitespace(pb)

        return body

    def parse_statement(self, pb):
        eat_whitespace(pb)
        pos = pb.lines
        name = parse_string(pb) if pb.peek() in ("'", '"') else self.parse_bare(pb)
        attrs = self.parse_attrs(pb)
        if pb.peek() == "{":
            body = self.parse_section_body(pb)
            el = Section(name=name, attrs=attrs, children=body, ctx=self.ctx)
        else:
            el = Directive(name=name, attrs=attrs, ctx=self.ctx)
        el.pos = pos
        eat_whitespace(pb)
        return el

    def parse_doc(self, pb):
        results = []
        while True:
            try:
                results.append(self.parse_statement(pb))
            except StopIteration:
                break
        return Root(children=results, ctx=self.ctx)


def parse_doc(f, ctx=None, line_end=";"):
    return DocParser(ctx, line_end=line_end).parse_doc(LineCounter(f))



from insights.tests import context_wrap
from insights.parsers.crypto_policies import CryptoPoliciesOpensshserver

CPSSHD_1 = """
CRYPTO_POLICY='-oCiphers=aes256-gcm@openssh.com,3des-cbc -oMACs=umac-128-etm@openssh.com'
""".strip()

CPSSHD_2 = """

CRYPTO_POLICY='-oCiphers=aes256-gcm@openssh.com,3des-cbc -oMACs=umac-128-etm@openssh.com,hmac-sha1-etm@openssh.com -oGSSAPIKexAlgorithms=gss-gex-sha1-,gss-group14-sha1- -oKexAlgorithms=curve25519-sha256@libssh.org,ecdh-sha2-nistp256,ecdh-sha2-nistp384,ecdh-sha2-nistp521,diffie-hellman-group-exchange-sha256,diffie-hellman-group14-sha256,diffie-hellman-group16-sha512,diffie-hellman-group18-sha512,diffie-hellman-group-exchange-sha1,diffie-hellman-group14-sha1 -oHostKeyAlgorithms=ssh-rsa,ssh-rsa-cert-v01@openssh.com,ssh-dss,ssh-dss-cert-v01@openssh.com,rsa-sha2-256,ecdsa-sha2-nistp256,ecdsa-sha2-nistp256-cert-v01@openssh.com,ecdsa-sha2-nistp384,ecdsa-sha2-nistp384-cert-v01@openssh.com,rsa-sha2-512,ecdsa-sha2-nistp521,ecdsa-sha2-nistp521-cert-v01@openssh.com,ssh-ed25519,ssh-ed25519-cert-v01@openssh.com -oPubkeyAcceptedKeyTypes=ssh-rsa,ssh-rsa-cert-v01@openssh.com,ssh-dss,ssh-dss-cert-v01@openssh.com,rsa-sha2-256,ecdsa-sha2-nistp256,ecdsa-sha2-nistp256-cert-v01@openssh.com,ecdsa-sha2-nistp384,ecdsa-sha2-nistp384-cert-v01@openssh.com,rsa-sha2-512,ecdsa-sha2-nistp521,ecdsa-sha2-nistp521-cert-v01@openssh.com,ssh-ed25519,ssh-ed25519-cert-v01@openssh.com'
"""  # no strip on purpose


def test_crypto_policies_opensshserver_1():
    result = CryptoPoliciesOpensshserver(context_wrap(CPSSHD_1))
    assert result["CRYPTO_POLICY"] == '-oCiphers=aes256-gcm@openssh.com,3des-cbc -oMACs=umac-128-etm@openssh.com'
    assert result.get("CRYPTO_POLICY") == '-oCiphers=aes256-gcm@openssh.com,3des-cbc -oMACs=umac-128-etm@openssh.com'
    assert result.get("OPTIONS1") is None
    assert "OPTIONS1" not in result
    assert "CRYPTO_POLICY" in result
    assert result.options == '-oCiphers=aes256-gcm@openssh.com,3des-cbc -oMACs=umac-128-etm@openssh.com'


def test_crypto_policies_opensshserver_2():
    result = CryptoPoliciesOpensshserver(context_wrap(CPSSHD_2))
    assert result["CRYPTO_POLICY"] == '-oCiphers=aes256-gcm@openssh.com,3des-cbc -oMACs=umac-128-etm@openssh.com,hmac-sha1-etm@openssh.com -oGSSAPIKexAlgorithms=gss-gex-sha1-,gss-group14-sha1- -oKexAlgorithms=curve25519-sha256@libssh.org,ecdh-sha2-nistp256,ecdh-sha2-nistp384,ecdh-sha2-nistp521,diffie-hellman-group-exchange-sha256,diffie-hellman-group14-sha256,diffie-hellman-group16-sha512,diffie-hellman-group18-sha512,diffie-hellman-group-exchange-sha1,diffie-hellman-group14-sha1 -oHostKeyAlgorithms=ssh-rsa,ssh-rsa-cert-v01@openssh.com,ssh-dss,ssh-dss-cert-v01@openssh.com,rsa-sha2-256,ecdsa-sha2-nistp256,ecdsa-sha2-nistp256-cert-v01@openssh.com,ecdsa-sha2-nistp384,ecdsa-sha2-nistp384-cert-v01@openssh.com,rsa-sha2-512,ecdsa-sha2-nistp521,ecdsa-sha2-nistp521-cert-v01@openssh.com,ssh-ed25519,ssh-ed25519-cert-v01@openssh.com -oPubkeyAcceptedKeyTypes=ssh-rsa,ssh-rsa-cert-v01@openssh.com,ssh-dss,ssh-dss-cert-v01@openssh.com,rsa-sha2-256,ecdsa-sha2-nistp256,ecdsa-sha2-nistp256-cert-v01@openssh.com,ecdsa-sha2-nistp384,ecdsa-sha2-nistp384-cert-v01@openssh.com,rsa-sha2-512,ecdsa-sha2-nistp521,ecdsa-sha2-nistp521-cert-v01@openssh.com,ssh-ed25519,ssh-ed25519-cert-v01@openssh.com'
    assert result.get("CRYPTO_POLICY") == '-oCiphers=aes256-gcm@openssh.com,3des-cbc -oMACs=umac-128-etm@openssh.com,hmac-sha1-etm@openssh.com -oGSSAPIKexAlgorithms=gss-gex-sha1-,gss-group14-sha1- -oKexAlgorithms=curve25519-sha256@libssh.org,ecdh-sha2-nistp256,ecdh-sha2-nistp384,ecdh-sha2-nistp521,diffie-hellman-group-exchange-sha256,diffie-hellman-group14-sha256,diffie-hellman-group16-sha512,diffie-hellman-group18-sha512,diffie-hellman-group-exchange-sha1,diffie-hellman-group14-sha1 -oHostKeyAlgorithms=ssh-rsa,ssh-rsa-cert-v01@openssh.com,ssh-dss,ssh-dss-cert-v01@openssh.com,rsa-sha2-256,ecdsa-sha2-nistp256,ecdsa-sha2-nistp256-cert-v01@openssh.com,ecdsa-sha2-nistp384,ecdsa-sha2-nistp384-cert-v01@openssh.com,rsa-sha2-512,ecdsa-sha2-nistp521,ecdsa-sha2-nistp521-cert-v01@openssh.com,ssh-ed25519,ssh-ed25519-cert-v01@openssh.com -oPubkeyAcceptedKeyTypes=ssh-rsa,ssh-rsa-cert-v01@openssh.com,ssh-dss,ssh-dss-cert-v01@openssh.com,rsa-sha2-256,ecdsa-sha2-nistp256,ecdsa-sha2-nistp256-cert-v01@openssh.com,ecdsa-sha2-nistp384,ecdsa-sha2-nistp384-cert-v01@openssh.com,rsa-sha2-512,ecdsa-sha2-nistp521,ecdsa-sha2-nistp521-cert-v01@openssh.com,ssh-ed25519,ssh-ed25519-cert-v01@openssh.com'
    assert result.get("OPTIONS1") is None
    assert "OPTIONS1" not in result
    assert "CRYPTO_POLICY" in result
    assert result.options == '-oCiphers=aes256-gcm@openssh.com,3des-cbc -oMACs=umac-128-etm@openssh.com,hmac-sha1-etm@openssh.com -oGSSAPIKexAlgorithms=gss-gex-sha1-,gss-group14-sha1- -oKexAlgorithms=curve25519-sha256@libssh.org,ecdh-sha2-nistp256,ecdh-sha2-nistp384,ecdh-sha2-nistp521,diffie-hellman-group-exchange-sha256,diffie-hellman-group14-sha256,diffie-hellman-group16-sha512,diffie-hellman-group18-sha512,diffie-hellman-group-exchange-sha1,diffie-hellman-group14-sha1 -oHostKeyAlgorithms=ssh-rsa,ssh-rsa-cert-v01@openssh.com,ssh-dss,ssh-dss-cert-v01@openssh.com,rsa-sha2-256,ecdsa-sha2-nistp256,ecdsa-sha2-nistp256-cert-v01@openssh.com,ecdsa-sha2-nistp384,ecdsa-sha2-nistp384-cert-v01@openssh.com,rsa-sha2-512,ecdsa-sha2-nistp521,ecdsa-sha2-nistp521-cert-v01@openssh.com,ssh-ed25519,ssh-ed25519-cert-v01@openssh.com -oPubkeyAcceptedKeyTypes=ssh-rsa,ssh-rsa-cert-v01@openssh.com,ssh-dss,ssh-dss-cert-v01@openssh.com,rsa-sha2-256,ecdsa-sha2-nistp256,ecdsa-sha2-nistp256-cert-v01@openssh.com,ecdsa-sha2-nistp384,ecdsa-sha2-nistp384-cert-v01@openssh.com,rsa-sha2-512,ecdsa-sha2-nistp521,ecdsa-sha2-nistp521-cert-v01@openssh.com,ssh-ed25519,ssh-ed25519-cert-v01@openssh.com'


def test_crypto_policies_opensshserver_empty():
    result = CryptoPoliciesOpensshserver(context_wrap(""))
    assert "OPTIONS1" not in result
    assert "CRYPTO_POLICY" not in result
    assert result.options is None



"""
Uname - command ``uname -a``
============================

The ``Uname`` class reads the output of the ``uname -a`` command and
interprets it.  It also does a number of handy extra things, like deriving
the RHEL release from the kernel version.

Uname objects can also be compared by their kernel versions.

An example from the following ``uname -a`` output::

    Linux server1.example.com 2.6.32-504.el6.x86_64 #1 SMP Tue Sep 16 01:56:35 EDT 2014 x86_64 x86_64 x86_64 GNU/Linux

Example:

    >>> type(uname)
    <class 'insights.parsers.uname.Uname'>
    >>> uname.version
    '2.6.32'
    >>> uname.release
    '504.el6'
    >>> uname.arch
    'x86_64'
    >>> uname.nodename
    'server1.example.com'

Uname objects can be created from, and compared to, other Uname objects or
kernel strings::

    >>> early_rhel6 = Uname.from_kernel('2.6.32-71')
    >>> late_rhel6 = Uname.from_release('6.7')
    >>> late_rhel6 > early_rhel6
    True
    >>> early_rhel6 > '2.6.32-279.el6.x86_64'
    False

"""

from collections import namedtuple
from distutils.version import LooseVersion, StrictVersion
from .. import parser, CommandParser
from insights.core.context import Context
from insights.specs import Specs

rhel_release_map = {
    "2.4.21-4": "3.0",
    "2.4.21-9": "3.1",
    "2.4.21-15": "3.2",
    "2.4.21-20": "3.3",
    "2.4.21-27": "3.4",
    "2.4.21-32": "3.5",
    "2.4.21-37": "3.6",
    "2.4.21-40": "3.7",
    "2.4.21-47": "3.8",
    "2.4.21-50": "3.9",
    "2.6.9-5": "4.0",
    "2.6.9-11": "4.1",
    "2.6.9-22": "4.2",
    "2.6.9-34": "4.3",
    "2.6.9-42": "4.4",
    "2.6.9-55": "4.5",
    "2.6.9-67": "4.6",
    "2.6.9-78": "4.7",
    "2.6.9-89": "4.8",
    "2.6.9-100": "4.9",
    "2.6.18-8": "5.0",
    "2.6.18-53": "5.1",
    "2.6.18-92": "5.2",
    "2.6.18-128": "5.3",
    "2.6.18-164": "5.4",
    "2.6.18-194": "5.5",
    "2.6.18-238": "5.6",
    "2.6.18-274": "5.7",
    "2.6.18-308": "5.8",
    "2.6.18-348": "5.9",
    "2.6.18-371": "5.10",
    "2.6.18-398": "5.11",
    "2.6.18-400": "5.11",
    "2.6.18-402": "5.11",
    "2.6.18-404": "5.11",
    "2.6.18-408": "5.11",
    "2.6.32-71": "6.0",
    "2.6.32-131": "6.1",
    "2.6.32-220": "6.2",
    "2.6.32-279": "6.3",
    "2.6.32-358": "6.4",
    "2.6.32-431": "6.5",
    "2.6.32-504": "6.6",
    "2.6.32-573": "6.7",
    "2.6.32-642": "6.8",
    "2.6.32-696": "6.9",
    "2.6.32-754": "6.10",
    "3.10.0-123": "7.0",
    "3.10.0-229": "7.1",
    "3.10.0-327": "7.2",
    "3.10.0-514": "7.3",
    "3.10.0-693": "7.4",
    "3.10.0-862": "7.5",
    "3.10.0-957": "7.6",
    "4.18.0-80": "8.0"
}

release_to_kernel_map = dict((v, k) for k, v in rhel_release_map.items())
RedhatRelease = namedtuple("RedhatRelease", ("major", "minor"))


class UnameError(Exception):
    """
    Exception subclass for errors related to uname content data and the
    Uname class.

    This exception should not be caught by rules plugins unless it is necessary
    for the plugin to return a particular answer when a problem occurs with
    uname data.  If a plugin catches this exception it must reraise it so that
    the engine has the opportunity to handle it/log it as necessary.
    """

    def __init__(self, msg, uname_line):
        """Class constructor

        :Parameters:
            - `msg`: Specific description of the error that occurred.
            - `uname_line`: Uname content that was being evaluated with the error
              occurred.
        """
        super(UnameError, self).__init__(msg, uname_line)
        self.msg = msg
        self.uname_line = uname_line

    def __str__(self):
        return "{0}:'{1}'".format(self.msg, self.uname_line)


@parser(Specs.uname)
class Uname(CommandParser):
    """
    A utility class to parse uname content data and compare version and release
    information.

    The input is a uname content string.  The content is parsed into specific
    uname elements that are made available through instance variables.
    Operators are provided for comparison of version and release information.
    Uname content is expected to be in the format returned by the ``uname -a``
    command.  The following instance variables are provided by this class:

    - `kernel`: Provides an unparsed copy of the full version and release
      string provided in the uname content input.  No validation is performed
      on this information. Generally in the format ``#.#.#-#.#.#.el#.arch``.
    - `name`: The kernel name, usually ``Linux``.
    - `nodename`: Hostname of the computer where the uname command was
      executed.  This information may obfuscated for security.
    - `version`: The major identification number for the kernel release. It
      should be in the format ``#.#.#[.#]`` or a UnameError exception will be
      raised.
    - `release`: The minor identification number for the kernel release. This
      information is generally in the format #.#.#.el#, however this is not
      strictly enforced.  If the release.arch information cannot be reliably
      parsed then `release` and `release_arch` will be the same value.
    - `release_arch`: This is the `release` plus the kernel architecture
      information as provided in `arch`.
    - `arch`: This contains the kernel architecture information like ``x86_64``
      or `s390`.  A list of known architectures is provided by the global
      variable `KNOWN_ARCHITECTURES`.  This information is not always
      present in the uname content.
    - `ver_rel`: This is a combination of `version` and `release` in the format
      ``version-release``.
    - `rhel_release`: A list of two elements, the major and minor RHEL product
      release numbers.

    """
    keys = [
        'os',
        'hw_platform',
        'processor',
        'machine',
        'kernel_date',
        'kernel_type',
        'kernel',
        'name',
        'nodename',
        'version',
        'release',
        'release_arch',
        'arch',
        'ver_rel',
        'rhel_release',
        '_lv_release',
        '_rel_maj',
        '_sv_version',
        '_lv_version'
    ]

    defaults = dict.fromkeys(keys)

    def __init__(self, context):
        super(Uname, self).__init__(context)
        if not context:
            return
        data = dict(self.defaults)
        data.update(self.data)
        self.data = data
        for k, v in self.data.items():
            setattr(self, k, v)

    def parse_content(self, content):
        """
        Parses uname content into individual uname components.

        :Parameters:
            - `content`: Uname content from Insights to be parsed.

        :Exceptions:
            - `UnameError`: Raised when there are any errors evaluating
              the uname content.
        """
        # Remove extra whitespace prior to parsing, preserve original line
        if not content:
            raise UnameError('Empty uname line', '')

        data = {}
        uname_list = [line for line in content if line.startswith("Linux")]
        if not uname_list:
            raise UnameError("Uname string appears invalid", '')
        uname_line = uname_list[0]
        uname_parts = uname_line.split(' ')
        if len(uname_parts) < 3:
            raise UnameError("Uname string appears invalid", uname_line)
        else:
            data['kernel'] = uname_parts[2]
            data['name'] = uname_parts[0]
            data['nodename'] = uname_parts[1]
        has_arch = "el" not in data['kernel'].split(".")[-1]
        try:
            data = self.parse_nvr(data['kernel'], data, arch=has_arch)
        except UnameError as error:
            error.uname_line = uname_line
            raise error

        # Additional uname content parsing, may not be as reliable
        if len(uname_parts) >= 15:
            data['os'] = uname_parts[-1]
            data['hw_platform'] = uname_parts[-2]
            data['processor'] = uname_parts[-3]
            data['machine'] = uname_parts[-4]
            data['kernel_date'] = " ".join(uname_parts[-10:-4])
            data['kernel_type'] = uname_parts[-11]
        else:
            data['os'] = None
            data['hw_platform'] = None
            data['processor'] = None
            data['machine'] = None
            data['kernel_date'] = None
            data['kernel_type'] = None
        self.data = data

    @classmethod
    def from_kernel(cls, kernel):
        """
        Create a Uname object from a kernel NVR (e.g. '2.6.32-504.el6.x86_64').

        :Parameters:
            - `kernel` - the kernel version and release string.
        """
        data = cls.parse_nvr(kernel, arch=False)
        content = ["{name} {nodename} {kernel} {kernel_type} {kernel_date} {machine} {processor} {hw_platform} {os}".format(
            name="Linux",
            nodename=data['nodename'],
            kernel=kernel,
            kernel_type=data['kernel_type'],
            kernel_date=data['kernel_date'],
            machine=data['machine'],
            processor=data['processor'],
            hw_platform=data['hw_platform'],
            os=data['os'])]
        return cls(Context(content=content, path=None))

    @classmethod
    def from_uname_str(cls, uname_str):
        """
        Create a Uname object from a string containing the output of 'uname -a'.

        :Parameters:
            - `uname_str` - the string output of `uname -a`
        """
        if len(uname_str.split(' ')) == 1:
            uname_str = "Linux " + "hostname " + uname_str
        return cls(Context(content=[uname_str.strip()], path=None))

    @classmethod
    def from_release(cls, release):
        """
        Attempt to create a Uname object from a release (e.g. '7.2').

        This translates from the release to the kernel version for that
        release, and then uses that to generate a Uname object using the
        class `from_kernel` method.  If the release does not match a known
        release, it returns None.

        :Parameters:
            - `release`: RHEL release version.
        """
        if isinstance(release, tuple):
            release = '.'.join(release)
        nvr = release_to_kernel_map.get(release)
        # return cls(Context(content=[nvr], path=None))
        return cls.from_kernel(nvr)

    @classmethod
    def parse_nvr(cls, nvr, data=None, arch=True):
        """
        Called by `parse_uname_line` to separate the version, release and arch
        information.

        :Parameters:
            - `nvr`: Uname content to parse.
            - `arch`: Flag to indicate whether there is architecture
              information in the release.

        :Exceptions:
            - `UnameError`: Raised on errors in evaluating the uname content.
        """
        if len(nvr.split('-')) < 2:
            raise UnameError("Too few parts in the uname version-release", nvr)

        if data is None:
            data = dict(cls.defaults)

        data['version'], data['release_arch'] = nvr.split('-', 1)
        if arch:
            try:
                data['release'], data['arch'] = data['release_arch'].rsplit('.', 1)
            except ValueError:
                data['release'] = data['release_arch']
                data['arch'] = None
        else:
            data['release'] = data['release_arch']
            data['arch'] = None
        data['_rel_maj'] = data['release'].split(".")[0]
        rhel_key = "-".join([data['version'], data['_rel_maj']])
        rhel_release = rhel_release_map.get(rhel_key, '-1.-1')
        data['rhel_release'] = rhel_release.split('.')
        try:
            data['_sv_version'] = StrictVersion(data['version'])
        except ValueError:
            data['_sv_version'] = None
        # LooseVersion doesn't raise errors, it just goes with what it gets
        data['_lv_version'] = LooseVersion(data['version'])
        num_sections = 7 if 'rt' in data['release'] else 5
        data['_lv_release'] = LooseVersion(pad_release(data['release'], num_sections=num_sections))
        data['ver_rel'] = '%s-%s' % (data['version'], data['release'])
        return data

    def _best_version(self, other):
        """
        Use strict versions for comparisons, or loose versions if strict are
        not available.
        """
        if self._sv_version and other._sv_version:
            return self._sv_version, other._sv_version
        else:
            return self._lv_version, other._lv_version

    def __str__(self):
        return "version: %s; release: %s; rel_maj: %s; lv_release: %s" % (
            self.version, self.release, self._rel_maj, self._lv_release
        )

    def __repr__(self):
        return "<%s '%s'>" % (self.__class__.__name__, str(self))

    def __eq__(self, other):
        """
        Operator to perform equal comparison of a Uname object to another Uname
        object or a string.

        The overloaded operators ``=``,``!=``,``<``,``>``,``<=``, and ``>=``
        provide logical comparison for a Uname object.  If the `other` object
        is a Uname or a string then comparison will be performed.  If `other`
        is any other type then currently a `NotImplementedError` exception is
        raised.

        Comparison is performed on the version and release information.
        Version comparison is strict, meaning that a the Uname class will only
        accept three integers as a version number such as "3.10.0".  Release
        comparison is also strict in that Uname expect a set of three numbers
        and a RHEL product string, however less than three numbers is accepted
        as long as there is at least one number.  For example "327.10.1.el7",
        "327.10.el7", and "327.el7" are all valid releases.  Also, for
        comparison purposes all missing numbers are assumed to be "0".
        For example:
            "327.10.0.el7" == "327.10.el7" and "327.0.0.el7" == "327.el7".

        :Parameters:
            - `other`: Uname object or version-release string to compare.

        :Exceptions:
            - `UnameError`: This exception is raised when there are any errors
              evaluating the uname content.
            - `ValueError`: This exception is raised when a version-release
              string is in an unparsable format.
            - `NotImplementedError`: This exception is thrown if a type other
              than ``Uname`` or ``str`` is provided as a parameter.
        """
        if isinstance(other, Uname):
            other_uname = other
        else:
            other_uname = Uname.from_uname_str(other)

        s_version, o_version = self._best_version(other_uname)

        return s_version == o_version and self._lv_release == other_uname._lv_release

    def __ne__(self, other):
        """
        Operator to perform inequality comparison of a Uname object to another
        Uname object or a string.

        See the `__eq__` operator for a detailed description.
        """
        return not(self == other)

    def __lt__(self, other):
        """
        Operator to perform less than comparison of a Uname object to another
        Uname object or a string.

        See the `__eq__` operator for a detailed description.
        """
        if isinstance(other, Uname):
            other_uname = other
        else:
            other_uname = Uname.from_uname_str(other)

        s_version, o_version = self._best_version(other_uname)

        return s_version < o_version or \
            (s_version == o_version and self._lv_release < other_uname._lv_release)

    def __le__(self, other):
        """
        Operator to perform less than or equal comparison of a Uname object
        to another Uname object or a string.

        See the :__eq__: operator for a detailed description.
        """
        return self < other or self == other

    def __gt__(self, other):
        """
        Operator to perform greater than comparison of a Uname object to
        another Uname object or a string.

        See the :__eq__: operator for a detailed description.
        """
        if isinstance(other, Uname):
            other_uname = other
        else:
            other_uname = Uname.from_uname_str(other)

        s_version, o_version = self._best_version(other_uname)

        return s_version > o_version or \
            (s_version == o_version and self._lv_release > other_uname._lv_release)

    def __ge__(self, other):
        """
        Operator to perform greater than or equal comparison of a Uname object
        to another Uname object or a string.

        See the :__eq__: operator for a detailed description.
        """
        return self.__gt__(other) or self.__eq__(other)

    def _less_than(self, other):
        """
        Compare two ``Uname`` classes.  This function is used like one might
        use ``__lt__``. However, if an invalid ``_lv_release`` indicated by a
        value of ``None``, ``False`` is returned.

        Since this behavior is not optimal for the the '<' comparison
        operator (raising an Error would probably be better) we'll keep it
        internal to the class.
        """
        s_version, o_version = self._best_version(other)

        if s_version == o_version:
            ret = self._lv_release < other._lv_release
        else:
            ret = s_version < o_version
        return ret

    def fixed_by(self, *fixes, **kwargs):
        """
        Determine whether the Uname object is fixed by a range of releases or
        by a specific release.

        :Parameters:
            - `fixes`: List of one or more Uname objects to compare to the
              current object.  `fixes` is a list of one or more Uname objects
              and each will be compared with the current object to determine
              a match.
            - `kwargs`: List of key word argument/Uname object pairs.
              Currently only `introduced_in` is supported as a keyword.  When
              used the current Uname object is checked to see if it is prior to
              the `introduced_in` release.  It will be further checked against
              fixes only if it is the same as or newer than the
              `introduced_in` release.
        """
        introduced_in = kwargs.get("introduced_in")
        if introduced_in and self._less_than(self.from_kernel(introduced_in)):
            return []
        fix_unames = sorted((self.from_kernel(f) for f in fixes))
        fix_kernels = [f.kernel for f in fix_unames]

        # If there is a fix in your kernel release stream
        # See if you are fixed by it
        for i, fixed in enumerate(fix_unames):
            if self._rel_maj == fixed._rel_maj:
                if self._less_than(fixed):
                    return fix_kernels[i:]
                else:
                    return []

        # No fixes for your specific release so just return
        # all fixes that are greater
        return [fix.kernel for fix in fix_unames if self._less_than(fix)]

    @property
    def release_tuple(self):
        return tuple(map(int, self.data["rhel_release"]))

    @property
    def redhat_release(self):
        return RedhatRelease(major=int(self.rhel_release[0]), minor=int(self.rhel_release[1]))


def pad_release(release_to_pad, num_sections=4):
    '''
    Pad out package and kernel release versions so that
    ``LooseVersion`` comparisons will be correct.

    Release versions with less than num_sections will
    be padded in front of the last section with zeros.

    For example ::

        pad_release("390.el6", 4)

    will return ``390.0.0.el6`` and ::

        pad_release("390.11.el6", 4)

    will return ``390.11.0.el6``.

    If the number of sections of the release to be padded is
    greater than num_sections, a ``ValueError`` will be raised.
    '''
    parts = release_to_pad.split('.')

    if len(parts) > num_sections:
        raise ValueError("Too many sections encountered ({found} > {num} in release string {rel}".format(
            found=len(parts), num=num_sections, rel=release_to_pad
        ))

    pad_count = num_sections - len(parts)
    return ".".join(parts[:-1] + ['0'] * pad_count + parts[-1:])



"""
RHNConf - file ``/etc/rhn/rhn.conf``
====================================

"""

from insights import Parser, parser, LegacyItemAccess
from insights.parsers import get_active_lines, unsplit_lines
from insights.specs import Specs


@parser(Specs.rhn_conf)
class RHNConf(LegacyItemAccess, Parser):
    """
    Class to parse the configuration file ``rhn.conf``.

    The special feature of ``rhn.conf`` is that values can span multiple
    lines with each intermediate line ending with a comma.

    This parser uses the :class:`insights.core.LegacyItemAccess` mix-in to provide
    access to its data directly.

    Attributes:
        data (dict): A dictionary of values keyed by the configuration
            item.  Values spanning multiple lines are compacted together.
            Values that include a comma are turned into lists.


    Sample ``rhn.conf`` input::

        # Corporate gateway (hostname:PORT):
        server.satellite.http_proxy = corporate_gateway.example.com:8080
        server.satellite.http_proxy_username =
        server.satellite.http_proxy_password =
        traceback_mail = test@example.com, test@redhat.com

        web.default_taskmaster_tasks = RHN::Task::SessionCleanup,
                                       RHN::Task::ErrataQueue,
                                       RHN::Task::ErrataEngine,
                                       RHN::Task::DailySummary,
                                       RHN::Task::SummaryPopulation,
                                       RHN::Task::RHNProc,
                                       RHN::Task::PackageCleanup

    Examples:
        >>> conf = shared[RHNConf]
        >>> conf.data['server.satellite.http_proxy']  # Long form access
        'corporate_gateway.example.com:8080'
        >>> conf['server.satellite.http_proxy']  # Short form access
        'corporate_gateway.example.com:8080'
        >>> conf['traceback_mail']  # split into a list
        ['test@example.com', 'test@redhat.com']
        >>> conf['web.default_taskmaster_tasks'][3]  # Values can span multiple lines
        'RHN::Task::DailySummary'
    """

    def parse_content(self, content):
        self.data = {}
        for line in unsplit_lines(get_active_lines(content), ',', keep_cont_char=True):
            if '=' in line:
                k, v = [i.strip() for i in line.split('=', 1)]
                self.data[k] = [i.strip() for i in v.split(',')] if ',' in v else v



from insights.parsers.rabbitmq_log import RabbitMQStartupLog
from insights.parsers.rabbitmq_log import RabbitMQStartupErrLog
from insights.parsers.rabbitmq_log import RabbitMQLogs
from insights.tests import context_wrap

from datetime import datetime

STARTUP_LOG = """
starting file handle cache server                                     ...done
starting worker pool                                                  ...done
starting database                                                     ...done
starting empty DB check                                               ...done
starting exchange recovery                                            ...done
starting queue supervisor and queue recovery                          ...BOOT ERROR: FAILED
"""

STARTUP_ERR_LOG = """
Error: {node_start_failed,normal}

Crash dump was written to: erl_crash.dump
Kernel pid terminated (application_controller) ({application_start_failure,kernel,{shutdown,{kernel,start,[normal,[]]}}})
"""


def test_rabbitmq_startup_log():
    log = RabbitMQStartupLog(context_wrap(STARTUP_LOG))
    assert len(log.get('done')) == 5


def test_rabbitmq_start_err_log():
    log = RabbitMQStartupErrLog(context_wrap(STARTUP_ERR_LOG))
    assert len(log.get('Error')) == 1


RABBIT_MQ_LOG = """
=INFO REPORT==== 7-Jun-2015::03:42:13 ===
accepting AMQP connection <0.13548.17> (192.168.100.40:59815 -> 192.168.100.41:5672)

=ERROR REPORT==== 7-Jun-2015::03:42:28 ===
AMQP connection <0.13548.17> (running), channel 19793 - error:
{amqp_error,frame_error,
            "type 65, all octets = <<>>: {frame_too_large,1342177289,131064}",
            none}

=ERROR REPORT==== 7-Jun-2015::03:42:31 ===
closing AMQP connection <0.13548.17> (192.168.100.40:59815 -> 192.168.100.41:5672):
fatal_frame_error
""".strip()


def test_rabbitmq_log():
    # Note that this tests one file - normally shared[RabbitMQLogs] will
    # contain multiple parser objects.
    log = RabbitMQLogs(context_wrap(RABBIT_MQ_LOG, path='/var/log/rabbitmq/rabbit@queue.example.com.log'))
    assert len(log.get('AMQP')) == 3
    assert len(list(log.get_after(datetime(2015, 6, 7, 3, 42, 20)))) == 9



from ...tests import context_wrap
from ..systemd.unitfiles import UnitFiles

KDUMP_DISABLED_RHEL7 = """
UNIT FILE                                   STATE
kdump.service                               disabled
""".strip()

KDUMP_ENABLED_RHEL7 = """
UNIT FILE                                   STATE
kdump.service                               enabled
""".strip()

KDUMP_ENABLED_FOOTER_RHEL7 = """
UNIT FILE                                   STATE
kdump.service                               enabled
1 unit file listed.
""".strip()

KDUMP_BIG_TEST = """
UNIT FILE                                   STATE
kdump.service                               enabled
other.service                               disabled
test.service                                static

3 unit files listed.
""".strip()

UNIT_INVALID_VS_VALID = """
UNIT FILE                                   STATE
svca.service                                enabled
svcb.service                                masked
svcc.service                                somenonsense

3 unit files listed.
"""


def test_unitfiles():
    context = context_wrap(KDUMP_DISABLED_RHEL7)
    unitfiles = UnitFiles(context)
    assert not unitfiles.is_on('kdump.service')
    assert len(unitfiles.services) == 1
    assert len(unitfiles.parsed_lines) == 1

    context = context_wrap(KDUMP_ENABLED_RHEL7)
    unitfiles = UnitFiles(context)
    assert unitfiles.is_on('kdump.service')
    assert len(unitfiles.services) == 1
    assert len(unitfiles.parsed_lines) == 1

    context = context_wrap(KDUMP_ENABLED_RHEL7)
    unitfiles = UnitFiles(context)
    assert unitfiles.is_on('kdump.service')
    assert len(unitfiles.services) == 1
    assert len(unitfiles.parsed_lines) == 1

    context = context_wrap(KDUMP_BIG_TEST)
    unitfiles = UnitFiles(context)
    assert unitfiles.is_on('kdump.service')
    assert not unitfiles.is_on('other.service')
    assert unitfiles.is_on('test.service')
    assert len(unitfiles.services) == 3
    assert len(unitfiles.parsed_lines) == 3

    context = context_wrap(UNIT_INVALID_VS_VALID)
    unitfiles = UnitFiles(context)
    assert unitfiles.is_on('svca.service')
    assert 'svca.service' in unitfiles.services
    assert 'svca.service' in unitfiles.service_list
    assert 'svcb.service' in unitfiles.services
    assert 'svcb.service' in unitfiles.service_list
    assert 'svcc.service' not in unitfiles.services
    assert 'svcc.service' not in unitfiles.service_list
    assert True is unitfiles.is_on('svca.service')
    assert False is unitfiles.is_on('svcb.service')
    assert None is unitfiles.is_on('svcc.service')
    assert unitfiles.exists('svca.service')
    assert unitfiles.exists('svcb.service')
    assert not unitfiles.exists('svcc.service')



"""
Ceph status commands
====================

This module provides processing for the output of the following ceph related
commands with `-f json-pretty` parameter.

CephOsdDump - command ``ceph osd dump -f json-pretty``
------------------------------------------------------

CephOsdDf - command ``ceph osd df -f json-pretty``
--------------------------------------------------

CephS - command ``ceph -s -f json-pretty``
------------------------------------------

CephDfDetail - command ``ceph osd erasure-code-profile get default -f json-pretty``
-----------------------------------------------------------------------------------

CephHealthDetail - command ``ceph daemon {ceph_socket_files} config show``
--------------------------------------------------------------------------

CephECProfileGet - command ``ceph health detail -f json-pretty``
----------------------------------------------------------------

CephCfgInfo - command ``ceph df detail -f json-pretty``
-------------------------------------------------------

CephOsdTree - command ``ceph osd tree -f json-pretty``
------------------------------------------------------

All these parsers are based on a shared class which processes the JSON
information into a dictionary.
"""

from .. import JSONParser, parser, CommandParser
from insights.specs import Specs


@parser(Specs.ceph_osd_dump)
class CephOsdDump(CommandParser, JSONParser):
    """
    Class to parse the output of ``ceph osd dump -f json-pretty``.

    Examples:

    >>> ceph_osd_dump_content = '''
        ... {
        ...     "epoch": 210,
        ...     "fsid": "2734f9b5-2013-48c1-8e96-d31423444717",
        ...     "created": "2016-11-12 16:08:46.307206",
        ...     "modified": "2017-03-07 08:55:53.301911",
        ...     "flags": "sortbitwise",
        ...     "cluster_snapshot": "",
        ...     "pool_max": 12,
        ...     "max_osd": 8,
        ...     "pools": [
        ...         {
        ...             "pool": 0,
        ...             "pool_name": "rbd",
        ...             "flags": 1,
        ...             "flags_names": "hashpspool",
        ...             "type": 1,
        ...             "size": 3,
        ...             "min_size": 2,
        ...             "crush_ruleset": 0,
        ...             "object_hash": 2,
        ...             "pg_num": 256,
        ...         }
        ...     ]
        ... }
        ... '''.strip()

    >>> result = CephOsdDump(context_wrap(ceph_osd_dump_content)).data
    >>> result['pools'][0]['min_size']
    2
    """
    pass


@parser(Specs.ceph_osd_df)
class CephOsdDf(CommandParser, JSONParser):
    """
    Class to parse the output of ``ceph osd df -f json-pretty``.

    Examples:

    >>> ceph_osd_dump_content = '''
        {
            "nodes": [
                {
                    "id": 0,
                    "name": "osd.0",
                    "type": "osd",
                    "type_id": 0,
                    "crush_weight": 1.091095,
                    "depth": 2,
                    "reweight": 1.000000,
                    "kb": 1171539620,
                    "kb_used": 4048208,
                    "kb_avail": 1167491412,
                    "utilization": 0.345546,
                    "var": 1.189094,
                    "pgs": 945
                }
            ],
            "stray": [],
            "summary": {
                "total_kb": 8200777340,
                "total_kb_used": 23831128,
                "total_kb_avail": 8176946212,
                "average_utilization": 0.290596,
                "min_var": 0.803396,
                "max_var": 1.189094,
                "dev": 0.035843
            }
        }
    '''.strip()
    >>> result = CephOsdDf(context_wrap(ceph_osd_df_content)).data
    >>> result['nodes'][0]['pgs']
    945
    """
    pass


@parser(Specs.ceph_s)
class CephS(CommandParser, JSONParser):
    """
    Class to parse the output of ``ceph -s -f json-pretty``.

    Examples:

    >>> ceph_s_content = '''
        {
            "health": {
                "health": {
                    "health_services": [
                        {
                            "mons": [
                                {
                                    "name": "dell-per430-22",
                                    "kb_total": 209612800,
                                    "kb_used": 3718628,
                                    "kb_avail": 205894172,
                                    "avail_percent": 98,
                                    "last_updated": "2017-03-10 22:38:25.920794",
                                    "store_stats": {
                                        "bytes_total": 36402550,
                                        "bytes_sst": 17659238,
                                        "bytes_log": 983040,
                                        "bytes_misc": 17760272,
                                        "last_updated": "0.000000"
                                    },
                                    "health": "HEALTH_OK"
                                }]
                            }
                        ]
                    }
                }
            },
            "pgmap": {
                "pgs_by_state": [
                    {
                        "state_name": "active+clean",
                        "count": 1800
                    }
                ],
                "version": 314179,
                "num_pgs": 1800,
                "data_bytes": 7943926574,
                "bytes_used": 24405610496,
                "bytes_avail": 8373190385664
                "bytes_total": 8397595996160
            }
        }
    '''.strip()
    >>> result = CephS(context_wrap(ceph_s_content)).data
    >>> result['pgmap']['pgs_by_state'][0]['state_name']
    'active+clean'
    """
    pass


@parser(Specs.ceph_df_detail)
class CephDfDetail(CommandParser, JSONParser):
    """
    Class to parse the output of ``ceph df detail -f json-pretty``.

    Examples:

    >>> ceph_df_content = '''
        {
            "stats": {
                "total_bytes": 17113243648,
                "total_used_bytes": 203120640,
                "total_avail_bytes": 16910123008,
                "total_objects": 0
            },
            "pools": [
                {
                    "name": "rbd",
                    "id": 0,
                    "stats": {
                        "kb_used": 0,
                        "bytes_used": 0,
                        "max_avail": 999252180,
                        "objects": 0,
                        "dirty": 0,
                        "rd": 0,
                        "rd_bytes": 0,
                        "wr": 0,
                        "wr_bytes": 0,
                        "raw_bytes_used": 0
                    }
                }
            ]
        }
    '''.strip()
    >>> result = CephDfDetail(context_wrap(ceph_df_content)).data
    >>> result['stats']['total_avail_bytes']
    16910123008
    """
    pass


@parser(Specs.ceph_health_detail)
class CephHealthDetail(CommandParser, JSONParser):
    """
    Class to parse the output of ``ceph health detail -f json-pretty``.

    Examples:

    >>> ceph_health_detail_content = '''
        {
            "health": {
            },
            "timechecks": {
                "epoch": 4,
                "round": 0,
                "round_status": "finished"
            },
            "summary": [],
            "overall_status": "HEALTH_OK",
            "detail": []
        }
    '''.strip()
    >>> result = CephHealthDetail(context_wrap(ceph_health_detail_content)).data
    >>> result["overall_status"]
    "HEALTH_OK"
    """
    pass


@parser(Specs.ceph_osd_ec_profile_get)
class CephECProfileGet(CommandParser, JSONParser):
    """
    Class to parse the output of ``ceph osd erasure-code-profile get default -f json-pretty``.

    Examples:

    >>> ceph_osd_ec_profile_get_content = '''
        {
            "k": "2",
            "m": "1",
            "plugin": "jerasure",
            "technique": "reed_sol_van"
        }
    '''.strip()
    >>> result = CephECProfileGet(context_wrap(ceph_osd_ec_profile_get_content)).data
    >>> result['k']
    "2"
    """
    pass


@parser(Specs.ceph_config_show)
class CephCfgInfo(CommandParser, JSONParser):
    """
    Class to parse the output of ``ceph daemon .. config show``

    Examples:

    >>> ceph_daemon_config_show = '''
        {
            "name": "osd.1",
            "cluster": "ceph",
            "debug_none": "0\/5",
            "heartbeat_interval": "5",
            "heartbeat_file": "",
            "heartbeat_inject_failure": "0",
            "perf": "true",
            "max_open_files": "131072",
            "ms_type": "simple",
            "ms_tcp_nodelay": "true",
            "ms_tcp_rcvbuf": "0",
            "ms_tcp_prefetch_max_size": "4096",
            "ms_initial_backoff": "0.2",
            "ms_max_backoff": "15",
            "ms_crc_data": "true",
            "ms_crc_header": "true",
            "ms_die_on_bad_msg": "false",
            "ms_die_on_unhandled_msg": "false",
            "ms_die_on_old_message": "false",
            "ms_die_on_skipped_message": "false",
            "ms_dispatch_throttle_bytes": "104857600",
            "ms_bind_ipv6": "false",
            "ms_bind_port_min": "6800",
            "ms_bind_port_max": "7300",
            "ms_bind_retry_count": "3",
            "ms_bind_retry_delay": "5"
        }
    '''.strip()
    >>> ceph_info = CephCfgInfo(context_wrap(ceph_daemon_config_show))
    >>> cpu_info.max_open_files
    '131072'
    """

    @property
    def max_open_files(self):
        """
        str: Return the value of max_open_files
        """
        return self.data["max_open_files"]


@parser(Specs.ceph_osd_tree)
class CephOsdTree(CommandParser, JSONParser):
    """
    Class to parse the output of the command "ceph osd tree -f json-pretty

    Examples:

    >>> ceph_osd_tree_content = '''
    {
    "nodes": [
        {
            "id": -1,
            "name": "default",
            "type": "root",
            "type_id": 10,
            "children": [
                -5,
                -4,
                -3,
                -2
            ]
        },
        {
            "id": -2,
            "name": "dhcp-192-56",
            "type": "host",
            "type_id": 1,
            "children": []
        },
        {
            "id": -3,
            "name": "dhcp-192-104",
            "type": "host",
            "type_id": 1,
            "children": []
        },
        {
            "id": -4,
            "name": "dhcp-192-67",
            "type": "host",
            "type_id": 1,
            "children": []
        },
        {
            "id": -5,
            "name": "localhost",
            "type": "host",
            "type_id": 1,
            "children": [
                1,
                3,
                5,
                2,
                4,
                0
            ]
        }
    ],
    "stray": []
    }
    '''.strip()
    >>> result = CephOsdTree(context_wrap(ceph_osd_tree_content))
    >>> result['nodes'][0]['children']
    ['-5', '-4', '-3', '-2']
    """
    pass
